import warnings

from pandas import options

verbose = True  # Controls the amount of feedback from the I-Server
fetch_on_init = True  # Controls if object will fetch basic data from server on init
progress_bar = True  # Controls whether progress bar will be shown during long fetch operations
debug = False  # Lets the program run in debugging mode
# Sets number of rows displayed for pandas DataFrame
options.display.max_rows = max(250, options.display.max_rows)
options.display.max_colwidth = max(100, options.display.max_colwidth)
# Warning settings: "error", "ignore", "always", "default", "module", "once"
print_warnings = 'always'
module_path = 'mstrio.*'
save_responses = False  # Used to save REST API responses for mocking
wip_warnings_enabled = True  # Controls whether warnings/errors about WIP functionality are emitted


def custom_formatwarning(msg, category, *args, **kwargs):
    # ignore everything except the message
    return str(category.__name__) + ': ' + str(msg) + '\n'


warnings.formatwarning = custom_formatwarning
warnings.filterwarnings(action=print_warnings, module=module_path)
warnings.filterwarnings(action=print_warnings, category=DeprecationWarning, module=module_path)
warnings.filterwarnings(action='default', category=UserWarning, module=module_path)



"""This is the demo script to show how to manage cube caches.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import Connection
from mstrio.project_objects import (OlapCube, CubeCache, list_cube_caches, delete_cube_caches,
                                    delete_cube_cache)

# get connection to an environment
base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
connection = Connection(base_url, username, password, project_name="MicroStrategy Tutorial",
                        login_mode=1)

# get caches from an OLAP Cube
oc_caches = OlapCube(connection, id='<cube id>').get_caches()

# list all caches
list_cube_caches(connection=connection)
# list all caches on a given node
list_cube_caches(connection=connection, nodes='<node name>')
# list all loaded caches on a given node
list_cube_caches(connection=connection, nodes='<node name>', loaded=True)
# list all cache on a given node for given cube
list_cube_caches(connection=connection, nodes='<node name>', cube_id='<cube id>')
# list all cache on a given node for given database connection
list_cube_caches(connection=connection, nodes='<node name>',
                 db_connection_id='<database connection id>')

# get a single cube cache by its id
cube_cache_ = CubeCache(connection=connection, cache_id='<cube cache id>')

# unload a cube cache
cube_cache_.unload()
# load a cube cache
cube_cache_.load()
# deactivate a cube cache
cube_cache_.deactivate()
# activate a cube cache
cube_cache_.activate()

# refresh cube cache
cube_cache_.fetch()

# get state of cube cache
cc_state = cube_cache_.state

# get properties of cube cache
cc_properties = cube_cache_.list_properties()

# Delete a single cube cache (there are two ways). When `force` argument is set
# to `False` (default value) then deletion must be approved.
cube_cache_.delete(force=True)
delete_cube_cache(connection=connection, id=cube_cache_.id, force=True)

# delete all cube caches (the same rule with `force` as above)
delete_cube_caches(connection=connection, force=True)
# delete all cube caches on a given node (the same rule with `force` as above)
delete_cube_caches(connection=connection, nodes='<node name>', force=True)
# delete all loaded cube caches on a given node (the same rule with `force` as
# above)
delete_cube_caches(connection=connection, nodes='<node name>', loaded=True, force=True)
# delete all cube caches on a given node for a given cube (the same rule with
# `force` as above)
delete_cube_caches(connection=connection, nodes='<node name>', cube_id='<cube id>', force=True)
# delete all cube caches on a given node for a given database connection (the
# same rule with `force` as above)
delete_cube_caches(connection=connection, nodes='<node name>',
                   db_connection_id='<database connection id>', force=True)



from mstrio import config
from mstrio.api import folders
from mstrio.object_management import PredefinedFolders
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import Entity, DeleteMixin, CopyMixin
from mstrio.utils.helper import fetch_objects_async, get_default_args_from_func

from typing import TYPE_CHECKING, Optional, List, Callable, Union

if TYPE_CHECKING:
    from mstrio.connection import Connection


def list_folders(connection: "Connection", project_id: Optional[str] = None,
                 to_dictionary: bool = False, limit: Optional[int] = None,
                 **filters) -> Union[List["Folder"], List[dict]]:
    """Get a list of folders - either all folders in a specific project or all
    folders that are outside of projects, called configuration-level folders.
    The list of configuration-level folders includes folders such as users, user
    groups, databases, etc. which are not project-specific. If you pass
    a `project_id`, you get folders in that project; if not, then you get
    configuration-level folders.

    Note:
        Id of project is not taken directly from `Connection` object, so you
        have to specify it explicitly.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (string): project ID
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg']

    Returns:
        list of `Folder` objects or list of dictionaries
    """
    objects = fetch_objects_async(connection, folders.list_folders, folders.list_folders_async,
                                  limit=limit, chunk_size=1000, project_id=project_id,
                                  filters=filters)

    if to_dictionary:
        return objects
    else:
        from mstrio.utils.object_mapping import map_objects_list
        return map_objects_list(connection, objects)


def get_my_personal_objects_contents(connection: "Connection", project_id: Optional[str] = None,
                                     to_dictionary: bool = False) -> List:
    """Get contents of `My Personal Objects` folder in a specific project.

    Note:
        When `project_id` is `None`, then its value is overwritten by
        `project_id` from `connection` object.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (string, optional): project ID
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.

    Returns:
        list of objects or list of dictionaries
    """
    objects = folders.get_my_personal_objects_contents(connection, project_id).json()
    if to_dictionary:
        return objects
    else:
        from mstrio.utils.object_mapping import map_objects_list
        return map_objects_list(connection, objects)


def get_predefined_folder_contents(connection: "Connection", folder_type: PredefinedFolders,
                                   project_id: Optional[str] = None, to_dictionary: bool = False,
                                   limit: Optional[int] = None, **filters) -> List:
    """Get contents of a pre-defined MicroStrategy folder in a specific project.
    Available values for `folder_type` are stored in enum `PredefinedFolders`.

    Note:
        When `project_id` is `None`, then its value is overwritten by
        `project_id` from `connection` object.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        folder_type (enum): pre-defined folder type. Available values are
            stored in enum `PredefinedFolders`.
        project_id (string, optional): project ID
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg']

    Returns:
        list of objects or list of dictionaries
    """
    objects = fetch_objects_async(connection, folders.get_predefined_folder_contents,
                                  folders.get_predefined_folder_contents_async, limit=limit,
                                  chunk_size=1000, folder_type=folder_type.value,
                                  project_id=project_id, filters=filters)

    if to_dictionary:
        return objects
    else:
        from mstrio.utils.object_mapping import map_objects_list
        return map_objects_list(connection, objects)


class Folder(Entity, DeleteMixin, CopyMixin):
    """Object representation of MicroStrategy Folder object.

    Attributes:
        connection: MSTR Connection object
        id: ID of the folder
        name: name of the folder
        description: description of the folder
        type: the type of the folder (based on EnumDSSXMLObjectTypes)
        subtype: the subtype of the folder (based on EnumDSSXMLObjectSubTypes)
        date_created: the date/time at which the folder was first saved into the
            metadata
        date_modified: the date/time at which the folder was last saved into the
            metadata
        version: the version number this folder is currently carrying
        acl: an array of access control entry objects
        owner: User object, the owner of the object
        acg: access rights, EnumDSSXMLAccessRightFlags
        ext_type: object extended type, EnumDSSExtendedType
        ancestors: list of ancestor folders
        abbreviation: folder's abbreviation
        hidden: specifies whether the folder is hidden
        icon_path: folder icon path
        view_media: view media settings
        certified_info: CertifiedInfo object, certification status, time of
            certification, and information about the certifier (currently only
            for document and report)
        contents: contents of folder
    """

    _FROM_DICT_MAP = {'owner': User.from_dict}

    _API_DELETE: Callable = staticmethod(folders.delete_folder)

    _OBJECT_TYPE = ObjectTypes.FOLDER
    _SIZE_LIMIT = 10000000  # this sets desired chunk size in bytes

    def __init__(self, connection: "Connection", id: str, name: Optional[str] = None):
        """Initialize folder object by its identifier.

        Note:
            Parameter `name` is not used when fetching. `id` is always used to
            uniquely identify folder.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str): Identifier of a pre-existing folder containing
                the required data.
            name (str): name of folder.
        """
        super().__init__(connection, id, name=name)

    @classmethod
    def create(cls, connection: "Connection", name: str, parent: str,
               description: Optional[str] = None) -> "Folder":
        """Create a new folder in a folder selected within connection object
        by providing its name, id of parent folder and optionally description.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            name (str): name of a new folder.
            parent (str): id of a parent folder in which new folder will be
                created.
            description (str, optional): optional description of a new folder.

        Returns:
            newly created folder
        """
        connection._validate_project_selected()
        response = folders.create_folder(connection, name, parent, description).json()
        if config.verbose:
            print("Successfully created folder named: '{}' with ID: '{}'".format(
                response.get('name'), response.get('id')))
        return cls.from_dict(source=response, connection=connection)

    def alter(self, name: Optional[str] = None, description: Optional[str] = None) -> None:
        """Alter the folder properties.

        Args:
            name: folder name
            description: folder description
        """
        func = self.alter
        default_dict = get_default_args_from_func(func)
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def get_contents(self, to_dictionary: bool = False, **filters) -> List:
        """Get contents of folder. It can contains other folders or different
        kinds of objects.

        Args:
            to_dictionary (bool, optional): If True returns dicts, by default
                (False) returns objects.
            **filters: Available filter parameters: ['id', 'name',
                'description', 'date_created', 'date_modified', 'acg']

        Returns:
            Contents as Python objects (when `to_dictionary` is `False` (default
            value)) or contents as dictionaries otherwise.
        """
        objects = fetch_objects_async(self.connection, folders.get_folder_contents,
                                      folders.get_folder_contents_async, limit=None,
                                      chunk_size=1000, id=self.id, filters=filters)

        if to_dictionary:
            return objects
        else:
            from mstrio.utils.object_mapping import map_objects_list
            return map_objects_list(self.connection, objects)



# flake8: noqa
from typing import Union

from .user import User, create_users_from_csv, list_users
from .user_group import UserGroup, list_user_groups
from .user_connections import UserConnections

UserOrGroup = Union[str, User, UserGroup]



# from .version import version_info, __version__


def _jupyter_nbextension_paths():
    return [dict(section="notebook",
                 src="production/mstr_jupyter/static",
                 dest="mstr_jupyter",
                 require="mstr_jupyter/main")]



"""This is the demo script to show how administrator can manage security
filters.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import Connection

from mstrio.access_and_security.security_filter import \
    list_security_filters, SecurityFilter, Qualification,\
    PredicateForm, PredicateFormFunction,\
    ConstantParameter, ConstantType,\
    AttributeRef, AttributeFormSystemRef

from mstrio.users_and_groups import list_users, list_user_groups

base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
conn = Connection(base_url, username, password, project_name="MicroStrategy Tutorial",
                  login_mode=1)

# list all security filters
list_security_filters(conn)

# initialize security filter by its id
sec_fil = SecurityFilter(conn, id='12345678123456781234567812345678')

# show information, qualification and members of the security filter
sec_fil.information.to_dict()
sec_fil.qualification.to_dict()
sec_fil.members

# create new security filter
name = "Year > 2015"
folder_id = "E28806A611E5B55F088C0080EF555BA1"
qualification = Qualification(
    tree=PredicateForm(function=PredicateFormFunction.GREATER, parameters=[
        ConstantParameter(ConstantType.DOUBLE, "2015.0")
    ], attribute=AttributeRef("8D679D5111D3E4981000E787EC6DE8A4", "Year"),
                       form=AttributeFormSystemRef("45C11FA478E745FEA08D781CEA190FE5", "ID")))

new_sec_fil = SecurityFilter.create(conn, qualification, name, folder_id)

# alter security filter - its name, description, qualification and folder
new_name = "Year > 2020"
new_description = "Year in this security filter was altered from 2015 to 2020"
# copy old qualification and change year
new_qualification = Qualification.from_dict(new_sec_fil.qualification.to_dict())
new_qualification.tree.parameters[0].constant['value'] = "2020.0"
new_folder_id = "E287FE0E11E5B55F03C70080EF555BA1"
sec_fil.alter(qualification=new_qualification, name=new_name, description=new_description)

# prepare users and user groups for applying and revoking security filter
mstr_user = list_users(conn, name_begins='MSTR U')[0]
system_monitors = list_user_groups(conn, name_begins='System Moni')[0]

# apply user(s) and/or user group(s) to security filter
sec_fil.apply([mstr_user, system_monitors.id])

# revoke user(s) and/or user group(s) from security filter
sec_fil.revoke([mstr_user.id, system_monitors])
sec_fil.revoke(sec_fil.members)  # revoke all members of security filter

# apply security filter directly to user or user group
mstr_user.apply_security_filter(sec_fil)

# revoke security filter directly from user or user group
system_monitors.apply_security_filter(sec_fil.id)

# fetch the latest security filter state
sec_fil.fetch(fetch_definition=False, fetch_members=True)

# delete security filter. When `force` argument is set
# to `False` (default value) then deletion must be approved.
sec_fil.delete(force=True)



import pandas as pd


class Model(object):
    """Internal utility for generating the definition of multi-table and
    single-table datasets.

    Create the definition of a super cube containing one or more tables. The
    definition includes the name and description of the super cube and the name
    and description of each table, attribute, and metric within the super cube.

    Attributes:
    """

    # dictionary key names
    _KEY_TABLE_NAME = 'table_name'
    _KEY_DATA_FRAME = 'data_frame'
    _KEY_AS_ATTR = 'to_attribute'
    _KEY_AS_METR = 'to_metric'
    _KEY_UPDATE_POL = 'update_policy'
    _KEY_TABLES = [_KEY_TABLE_NAME, _KEY_DATA_FRAME, _KEY_AS_ATTR, _KEY_AS_METR]

    __MAX_DESC_LEN = 250

    _INVALID_CHARS = ['\\', '"', '[', ']']  # check for invalid characters in column names

    def __init__(self, tables, name, description=None, folder_id=None, ignore_special_chars=False):
        """Initializes Model with tables, a name, and an optional description.

        Args:
            tables (list of dicts): List of one or more dictionaries with keys
                `table_name`, `data_frame`, and optionally `as_attribute` and
                `as_metric`. Note that `as_attribute` and `as_metric` should be
                used when the default Python data type (e.g. `int`) should be
                converted to an attribute instead of a metric in MicroStrategy.
            name (str): Name of the data set.
            description (str, optional): Description of the data set. Must be
                less than or equal to 250 characters.
            folder_id (str, optional): ID of the shared folder that the super
                cube should be created within. If `None`, defaults to the user's
                My Reports folder.
        """

        self.__ignore_special_chars = ignore_special_chars

        # check integrity of tables list
        self.__check_table_list(tables=tables)

        # check super cube name params
        self.__name = name
        self.__check_param_str(self.__name, msg="SuperCube name should be a string.")
        self.__check_param_len(
            self.__name,
            msg="SuperCube name should be <= {} characters.".format(self.__MAX_DESC_LEN),
            max_length=self.__MAX_DESC_LEN)
        self.__check_param_inv_chars(
            self.__name, msg="SuperCube name cannot contain '{}', '{}', '{}', '{}'.".format(
                *self._INVALID_CHARS), invalid_chars=self._INVALID_CHARS)

        # check super cube description params
        if description is None:
            self.__description = ""
        else:
            self.__description = description
            self.__check_param_str(self.__description,
                                   msg="SuperCube description should be a string.")
            self.__check_param_len(
                self.__description, msg="SuperCube description should be <= {} characters.".format(
                    self.__MAX_DESC_LEN), max_length=self.__MAX_DESC_LEN)

        # check folder_id param
        if folder_id is None:
            self._folder_id = ""
        else:
            self._folder_id = folder_id

        # init lists to accumulate table, attr, metric definitions and model
        self.__tables = []
        self.__attributes = []
        self.__metrics = []
        self.__model = None

        # build the model
        self.__build(tables=tables)

    def get_model(self):
        """Return the model object."""
        return self.__model

    def __build(self, tables):
        """Generates the data model by mapping attributes and metrics from list
        of tables."""

        for table in tables:

            # map column names and column types
            _col_names = self.__get_col_names(table[self._KEY_DATA_FRAME])
            _col_types = self.__get_col_types(table[self._KEY_DATA_FRAME])

            # map tables
            self.__add_table(name=table[self._KEY_TABLE_NAME], col_names=_col_names,
                             col_types=_col_types)

            # map attributes and metrics
            for _name, _type in zip(_col_names, _col_types):

                if self.__is_metric(_type):
                    if self._KEY_AS_ATTR in table.keys() and _name in table[self._KEY_AS_ATTR]:
                        self.__add_attribute(_name, table[self._KEY_TABLE_NAME])
                    else:
                        self.__add_metric(_name, table[self._KEY_TABLE_NAME])

                else:
                    if self._KEY_AS_METR in table.keys() and _name in table[self._KEY_AS_METR]:
                        self.__add_metric(_name, table[self._KEY_TABLE_NAME])
                    else:
                        self.__add_attribute(_name, table[self._KEY_TABLE_NAME])

        # set model object
        self.__model = {
            "name": self.__name,
            "description": self.__description,
            "folderId": self._folder_id,
            "tables": self.__tables,
            "metrics": self.__metrics,
            "attributes": self.__attributes
        }

    def __get_col_types(self, table):
        """Map column types from each column in the list of table."""
        return list(map(self.__map_data_type, list(table.dtypes.values)))

    def __add_metric(self, name, table_name):
        """Add a metric to a metric list instance."""
        self.__metrics.append({
            'name': name,
            'expressions': [{
                'tableName': table_name,
                'columnName': name
            }]
        })

    def __add_attribute(self, name, table_name):
        """Add an attribute to an attribute list instance."""
        self.__attributes.append({
            'name': name,
            'attributeForms': [{
                'category': 'ID',
                'expressions': [{
                    'tableName': table_name,
                    'columnName': name
                }]
            }]
        })

    def __add_table(self, name, col_names, col_types):
        """Add a table to a table list instance."""
        self.__tables.append({
            'name': name,
            'columnHeaders': [{
                'name': name,
                'dataType': typ
            } for name, typ in zip(col_names, col_types)]
        })

    def __check_table_list(self, tables):
        """Check integrity of table list parameter."""

        # tables must be a list
        if not isinstance(tables, list):
            raise TypeError("Elements of tables must be a list of dicts.")

        # tables cannot be length 0
        if len(tables) == 0:
            raise ValueError("No tables have been added to the super cube.")

        # check integrity of each table passed to tables
        for table in tables:
            self.__check_table(table)

    def __check_table(self, table):
        """Check integrity of table parameter."""

        # force all list elements to be a dict with specific names
        msg = "Each table must be a dictionary with keys: '{}', '{}', '{},' and '{}'.".format(
            *self._KEY_TABLES)
        if not isinstance(table, dict):
            raise TypeError(msg)

        if not all(k in table.keys() for k in (self._KEY_TABLE_NAME, self._KEY_DATA_FRAME)):
            raise ValueError(msg)

        # check that the value of the data frame key is a pandas data frame
        if not isinstance(table[self._KEY_DATA_FRAME], pd.DataFrame):
            msg = "Pandas DataFrame must be passed as the value in the '{}' key.".format(
                self._KEY_DATA_FRAME)
            raise TypeError(msg)

        # check for presence of invalid characters in data frame column names
        if not self.__ignore_special_chars and any([
                col for col in table[self._KEY_DATA_FRAME].columns
                for inv in self._INVALID_CHARS if inv in col
        ]):
            msg = "Column names cannot contain '{}', '{}', '{}', '{}'".format(*self._INVALID_CHARS)
            raise ValueError(msg)

    @staticmethod
    def __get_col_names(table):
        """Returns column names from a table as a list."""
        return list(table.columns)

    @staticmethod
    def __map_data_type(datatype):
        """Maps a Python data type to a MicroStrategy data type."""
        if datatype == 'object':
            return "STRING"
        elif datatype == 'int32':
            return "INTEGER"
        elif datatype == 'int64':
            return "BIGINTEGER"
        elif datatype in ['float64', 'float32']:
            return "DOUBLE"
        elif datatype == 'bool':
            return "BOOLEAN"
        elif datatype == 'datetime64[ns]':
            return 'DATETIME'

    @staticmethod
    def __check_param_len(param, msg, max_length):
        if len(param) > max_length:
            raise ValueError(msg)
        else:
            return True

    @staticmethod
    def __check_param_str(param, msg):
        if not isinstance(param, str):
            raise TypeError(msg)
        else:
            return True

    @staticmethod
    def __check_param_inv_chars(param, msg, invalid_chars):
        if any([inv for inv in invalid_chars if inv in param]):
            raise ValueError(msg)

    @staticmethod
    def __is_metric(datatype):
        """Helper function for determining if the requested datatype is (by
        default) a metric or attribute."""
        if datatype in ["DOUBLE", "INTEGER", "BIGINTEGER", "BIGDECIMAL"]:
            return True
        else:
            return False

    @property
    def name(self):
        return self.__name

    @property
    def description(self):
        return self.__description

    @property
    def folder_id(self):
        return self._folder_id

    @property
    def tables(self):
        return self.__tables

    @property
    def attributes(self):
        return self.__attributes

    @property
    def metrics(self):
        return self.__metrics



# flake8: noqa
from .cube import list_all_cubes, load_cube
from .cube_cache import CubeCache, delete_cube_cache, delete_cube_caches, list_cube_caches
from .olap_cube import list_olap_cubes, OlapCube
from .super_cube import list_super_cubes, SuperCube



from .objects_and_references import ObjectInformation, ObjectReference, AttributeRef,\
    AttributeFormSystemRef, AttributeFormNormalRef, AttributeElement, ElementPromptRef,\
    FilterRef  # noqa: F401

from .predicate_parameters import ParameterType, ParameterBase, ConstantType, ConstantParameter,\
    ObjectReferenceParameter, ExpressionParameter, PromptParameter, DynamicDateTimeParameter,\
    ConstantArrayParameter  # noqa: F401

from .predicates import PredicateBase, PredicateFormFunction, PredicateForm,\
    PredicateElementListFunction, PredicateElementList, PredicateFilter,\
    PredicateJointElementList, LogicFunction, LogicOperator  # noqa: F401

from .qualification import Qualification  # noqa: F401

from .security_filter import SecurityFilter, list_security_filters  # noqa: F401



"""This is the demo script to show how administrator can manage projects
and its settings.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import Connection
from mstrio.server import Project, compare_project_settings, Environment

base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
conn = Connection(base_url, username, password, project_name="MicroStrategy Tutorial",
                  login_mode=1)
env = Environment(connection=conn)

# get list of all projects or those just loaded
all_projects = env.list_projects()
loaded_projects = env.list_loaded_projects()

# get project with a given name
project = Project(connection=conn, name="MicroStrategy Tutorial")
# idle/resume an project on a given node(s) (name(s) of existing node(s)
# should be used)
project.idle(on_nodes=["node1"])
project.resume(on_nodes="node1")
# unload/load an project from/to a given node(s) (name(s) of existing
# node(s) should be used)
project.unload(on_nodes="node1")
project.load(on_nodes="node1")

# get settings of an project as a dataframe
project_settings_df = project.settings.to_dataframe

# create an project and store it in variable to have immediate access to it
new_project = env.create_project(name="Demo Project 1", description="some description")
# change description of a newly created project
new_project.alter(
    description="This project is used to present creation of a project as a demo")

# compare settings of 2 projects (only differences will be stored into
# a DataFrame)
project1 = Project(connection=conn, name="Consolidated Education Project")
project2 = Project(connection=conn, name="Platform Analytics")
df_cmp = compare_project_settings(projects=[project, project1, project2], show_diff_only=True)

# save/load settings of an project to/from a file (format can be 'csv',
# 'json' or 'pickle')
project.settings.to_csv(name="mstr_tutorial_settings.csv")
project.settings.import_from(file="mstr_tutorial_settings.csv")

# change a setting of an project
project.settings.cubeIndexGrowthUpperBound = 501
project.settings.update()



from datetime import datetime
from enum import Enum
from typing import List, Optional, Union

from mstrio.utils.helper import camel_to_snake, Dictable, exception_handler


class SendContentAs(Enum):
    DATA = "data"
    DATA_AND_HISTORY_LIST = "data_and_history_list"
    DATA_AND_LINK_AND_HISTORY_LIST = "data_and_link_and_history_list"
    LINK_AND_HISTORY_LIST = "link_and_history_list"


class Orientation(Enum):
    PORTRAIT = "PORTRAIT"
    LANDSCAPE = "LANDSCAPE"


class CacheType(Enum):
    RESERVED = "reserved"
    SHORTCUT = "shortcut"
    SHORTCUTWITHBOOKMARK = "shortcut_and_bookmark"


class ShortcutCacheFormat(Enum):
    RESERVED = "reserved"
    JSON = "json"
    BINARY = "binary"
    BOTH = "both"


class ClientType(Enum):
    RESERVED = "RESERVED"
    BLACKBERRY = "BLACKBERRY"
    PHONE = "PHONE"
    TABLET = "TABLET"
    ANDROID = "ANDROID"


class LibraryCacheTypes(Enum):
    ANDROID = "android"
    ANDROID_AND_IOS = "android_and_ios"
    IOS = "ios"
    WEB = "web"


class DeliveryDictable(Dictable):
    VALIDATION_DICT = {}

    @classmethod
    def from_dict(cls, source, **kwargs):
        """Initialize Delivery object from dictionary."""
        obj = cls.__new__(cls)
        super(DeliveryDictable, obj).__init__()
        source = camel_to_snake(source)
        for key, value in source.items():
            if key == 'zip':
                setattr(obj, key, ZipSettings.from_dict(value))
            else:
                setattr(obj, key, value)
        return obj

    def validate(self):
        for key, value in self.__dict__.items():
            vtype = self.VALIDATION_DICT[key][0]
            obligatory = self.VALIDATION_DICT[key][1]
            if value and not isinstance(value, vtype):
                exception_handler(
                    "{} has incorrect type {}. Correct type is {}.".format(
                        key, type(value), vtype), TypeError)
            elif value is None and obligatory:
                exception_handler("{} is obligatory and cannot be empty.".format(key), ValueError)


class ZipSettings(DeliveryDictable):
    """Optional compression settings

    Attributes:
        filename: Filename of the compressed content
        password: Optional password for the compressed file
        password_protect: Whether to password protect file or not
    """
    VALIDATION_DICT = {
        "filename": [str, False],
        "password": [str, False],
        "password_protect": [bool, False]
    }

    def __init__(self, filename: Optional[str] = None, password: Optional[str] = None,
                 password_protect: bool = False):
        self.filename = filename
        self.password = password if password_protect else None
        self.password_protect = password_protect
        self.validate()


class Delivery(DeliveryDictable):
    """Delivery settings object

    Attributes:
        mode: The subscription delivery mode (i.e. email, file, printer, etc.)
        expiration: Expiration date of the subscription, format should
            be yyyy-MM-dd
        contact_security: Whether to use contact security for each contact
            group member
        email: Email delivery properties object
        file: File delivery properties object
        printer: File delivery properties object
        ftp: FTP delivery properties object
        cache: Cache delivery properties
        mobile: Mobile delivery properties object
        history_list: HistoryList delivery properties
    """

    class DeliveryMode(Enum):
        EMAIL = "EMAIL"
        FILE = "FILE"
        PRINTER = "PRINTER"
        HISTORY_LIST = "HISTORY_LIST"
        CACHE = "CACHE"
        MOBILE = "MOBILE"
        FTP = "FTP",
        SNAPSHOT = "SNAPSHOT",
        PERSONAL_VIEW = "PERSONAL_VIEW"
        SHARED_LINK = "SHARED_LINK"
        UNSUPPORTED = "UNSUPPORTED"

    class Email(DeliveryDictable):
        """Delivery properties for Email subscriptions

        Attributes:
            subject: The email subject associated with the subscription
            message: The email body of subscription
            filename: The filename that will be delivered when the subscription
                is executed
            space_delimiter: The space delimiter
            send_content_as: Send subscribed content as one of [data,
                data_and_history_list, data_and_link_and_history_list,
                link_and_history_list]
            overwrite_older_version: Whether the current subscription will
                overwrite earlier versions of the same report or document
                in the history list
            zip: Optional compression settings object
        """
        VALIDATION_DICT = {
            "subject": [str, True],
            "message": [str, False],
            "filename": [str, False],
            "space_delimiter": [str, False],
            "send_content_as": [str, False],
            "overwrite_older_version": [bool, False],
            "zip": [ZipSettings, False]
        }

        def __init__(self, subject: Optional[str] = None, message: Optional[str] = None,
                     filename: Optional[str] = None, space_delimiter: Optional[str] = None,
                     send_content_as: Optional[SendContentAs] = None,
                     overwrite_older_version: bool = False, zip: Optional[ZipSettings] = None):
            self.subject = subject
            self.message = message
            self.filename = filename
            self.space_delimiter = space_delimiter
            self.send_content_as = send_content_as
            self.overwrite_older_version = overwrite_older_version
            self.zip = zip
            self.validate()

        def validate(self):
            if self.message and len(self.message) > 1000:
                exception_handler("Message too long. Max message length is 1000 characters.")
            if self.subject and len(self.subject) > 265:
                exception_handler("Subject too long. Max subject length is 265 characters.")
            super().validate()

    class File(DeliveryDictable):
        """Delivery properties for File subscriptions

        Attributes:
            filename: The filename that will be delivered when the subscription
                is executed
            space_delimiter: The space delimiter
            burst_sub_folder:The burst sub folder
            zip: Optional compression settings object
        """
        VALIDATION_DICT = {
            "filename": [str, False],
            "space_delimiter": [str, False],
            "burst_sub_folder": [str, False],
            "zip": [ZipSettings, False]
        }

        def __init__(self, filename: Optional[str] = None, space_delimiter: Optional[str] = None,
                     burst_sub_folder: Optional[str] = None, zip: Optional[ZipSettings] = None):
            self.filename = filename
            self.space_delimiter = space_delimiter
            self.burst_sub_folder = burst_sub_folder
            self.zip = zip

    class Printer(DeliveryDictable):
        """Delivery properties for Printer subscriptions

        Attributes:
            copies: The number of copies that should be printed
            range_start: The number indicating the first report page that
                should be printed
            range_end: The number indicating the last report page that should
                be printed
            collated: Whether the printing should be collated or not
            orientation: Whether orientation is portrait or landscape
            use_print_range: Whether a print range should be used
        """
        VALIDATION_DICT = {
            "copies": [int, False],
            "range_start": [int, False],
            "range_end": [int, False],
            "collated": [bool, False],
            "orientation": [str, False],
            "use_print_range": [bool, False]
        }

        def __init__(self, copies: Optional[int] = None, range_start: Optional[int] = None,
                     range_end: Optional[int] = None, collated: bool = False,
                     orientation: Orientation = Orientation.PORTRAIT.name,
                     use_print_range: bool = False):
            self.copies = copies
            self.range_start = range_start
            self.range_end = range_end
            self.collated = collated
            self.orientation = orientation
            self.use_print_range = use_print_range

    class Ftp(DeliveryDictable):
        """Delivery properties for FTP subscriptions

        Attributes:
            filename: The filename that will be delivered when the subscription
                is executed
            space_delimiter: The space delimiter
            zip: Optional compression settings object
        """
        VALIDATION_DICT = {
            "filename": [str, False],
            "space_delimiter": [str, False],
            "zip": [ZipSettings, False]
        }

        def __init__(self, space_delimiter: Optional[str] = None, filename: Optional[str] = None,
                     zip: Optional[ZipSettings] = None):
            self.space_delimiter = space_delimiter
            self.filename = filename
            self.zip = zip

    class Cache(DeliveryDictable):
        """Delivery properties for Cache subscriptions

        Attributes:
            cache_type: The cache type to use
            shortcut_cache_format: The shortcut cache format to use
            library_cache_types: Set of library cache types,
                available types can be web, android, ios
            reuse_dataset_cache: Whether to reuse dataset cache
            is_all_library_users: Whether for all library users
        """
        VALIDATION_DICT = {
            "library_cache_types": [list, False],
            "reuse_dataset_cache": [bool, False],
            "is_all_library_users": [bool, False],
        }

        def __init__(self, cache_type: CacheType = CacheType.RESERVED,
                     shortcut_cache_format: ShortcutCacheFormat = ShortcutCacheFormat.RESERVED,
                     library_cache_types: List[LibraryCacheTypes] = None,
                     reuse_dataset_cache: bool = False, is_all_library_users: bool = False):
            self.cache_type = cache_type
            self.shortcut_cache_format = shortcut_cache_format
            self.library_cache_types = library_cache_types
            self.reuse_dataset_cache = reuse_dataset_cache
            self.is_all_library_users = is_all_library_users

    class Mobile(DeliveryDictable):
        """Delivery properties for Mobile subscriptions

        Attributes:
            client_type: The mobile client type
            device_id: The mobile target project
            do_not_create_update_caches: Whether the subscription will use
                a existing cache
            overwrite_older_version: Whether the current subscription will
                overwrite earlier versions of the same report or document in
                the history list
            re_run_hl: Whether the subscription will re-run against warehouse
        """
        VALIDATION_DICT = {
            "client_type": [str, False],
            "device_id": [str, False],
            "do_not_create_update_caches": [bool, False],
            "overwrite_older_version": [bool, False],
            "re_run_hl": [bool, False]
        }

        def __init__(self, client_type: ClientType = ClientType.RESERVED.name,
                     device_id: Optional[str] = None, do_not_create_update_caches: bool = False,
                     overwrite_older_version: bool = False, re_run_hl: bool = False):
            self.client_type = client_type
            self.device_id = device_id
            self.do_not_create_update_caches = do_not_create_update_caches
            self.overwrite_older_version = overwrite_older_version
            self.re_run_hl = re_run_hl

    class HistoryList(DeliveryDictable):
        """Delivery properties for History List subscriptions

        Attributes:
            device_id: The mobile target project
            do_not_create_update_caches: Whether the subscription will use
                a existing cache
            overwrite_older_version: Whether the current subscription will
                overwrite earlier versions of the same report or document in
                the history list
            re_run_hl: Whether the subscription will re-run against warehouse
        """
        VALIDATION_DICT = {
            "device_id": [str, False],
            "do_not_create_update_caches": [bool, False],
            "overwrite_older_version": [bool, False],
            "re_run_hl": [bool, False]
        }

        def __init__(self, device_id: Optional[str] = None,
                     do_not_create_update_caches: bool = False,
                     overwrite_older_version: bool = False, re_run_hl: bool = False):
            self.device_id = device_id
            self.do_not_create_update_caches = do_not_create_update_caches
            self.overwrite_older_version = overwrite_older_version
            self.re_run_hl = re_run_hl

    VALIDATION_DICT = {
        "mode": [str, True],
        "expiration": [str, False],
        "contact_security": [bool, False],
        "email": [Email, False],
        "file": [File, False],
        "printer": [Printer, False],
        "ftp": [Ftp, False],
        "cache": [Cache, False],
        "mobile": [Mobile, False],
        "history_list": [HistoryList, False]
    }

    def __init__(self, mode=DeliveryMode.EMAIL.value, expiration: str = None,
                 contact_security: Optional[bool] = None, subject: Optional[str] = None,
                 message: Optional[str] = None, filename: Optional[str] = None,
                 compress: bool = False, zip: Optional[ZipSettings] = None,
                 password: Optional[str] = None, password_protect: bool = False,
                 space_delimiter: Optional[str] = None,
                 send_content_as: SendContentAs = SendContentAs.DATA,
                 overwrite_older_version: bool = False, burst_sub_folder: Optional[str] = None,
                 copies: Optional[int] = None, range_start: Optional[int] = None,
                 range_end: Optional[int] = None, collated: bool = False,
                 orientation: Orientation = Orientation.PORTRAIT.name,
                 use_print_range: bool = False, client_type: ClientType = ClientType.RESERVED.name,
                 device_id: Optional[str] = None, do_not_create_update_caches: bool = False,
                 re_run_hl: bool = False, email: Optional[Email] = None,
                 file: Optional[File] = None, cache_type: Union[CacheType,
                                                                str] = CacheType.RESERVED,
                 shortcut_cache_format: Union[ShortcutCacheFormat,
                                              str] = ShortcutCacheFormat.RESERVED,
                 library_cache_types: List[Union[LibraryCacheTypes, str]] = [LibraryCacheTypes.WEB],  # noqa: E501
                 reuse_dataset_cache: bool = False,
                 is_all_library_users: bool = False, notification_enabled: bool = False,
                 personal_notification_address_id: Optional[str] = None):
        self.mode = mode
        if expiration:
            if not self._expiration_check(expiration):
                exception_handler(
                    "Expiration date must be later or equal to {}".format(
                        datetime.now().strftime("%Y-%m-%d")), ValueError)
            else:
                self.expiration = expiration
        self.contact_security = contact_security
        temp_zip = ZipSettings(
            filename, password, password_protect
        ) if zip is None and compress else zip if zip is not None and compress else None

        if self.DeliveryMode(mode) == self.DeliveryMode.EMAIL:
            self.email = self.Email(subject, message, filename, space_delimiter, send_content_as,
                                    overwrite_older_version, temp_zip)
        elif self.DeliveryMode(mode) == self.DeliveryMode.FILE:
            self.file = self.File(filename, space_delimiter, burst_sub_folder, temp_zip)
        elif self.DeliveryMode(mode) == self.DeliveryMode.PRINTER:
            self.printer = self.Printer(copies, range_start, range_end, collated, orientation,
                                        use_print_range)
        elif self.DeliveryMode(mode) == self.DeliveryMode.FTP:
            self.ftp = self.Ftp(space_delimiter, filename, temp_zip)
        elif self.DeliveryMode(mode) == self.DeliveryMode.CACHE:
            self.cache = self.Cache(cache_type, shortcut_cache_format, library_cache_types,
                                    reuse_dataset_cache, is_all_library_users)
        elif self.DeliveryMode(mode) == self.DeliveryMode.MOBILE:
            self.mobile = self.Mobile(client_type, device_id, do_not_create_update_caches,
                                      overwrite_older_version, re_run_hl)
        elif self.DeliveryMode(mode) == self.DeliveryMode.HISTORY_LIST:
            self.history_list = self.HistoryList(device_id, do_not_create_update_caches,
                                                 overwrite_older_version, re_run_hl)
        if notification_enabled:
            self.notification_enabled = notification_enabled
            self.personal_notification = {"address_id": personal_notification_address_id}

    @classmethod
    def from_dict(cls, source, **kwargs):
        """Initialize Delivery object from dictionary."""
        obj = cls.__new__(cls)
        super(Delivery, obj).__init__()
        source = camel_to_snake(source)
        for key, value in source.items():
            if not obj.change_mode(key, value):  # Change mode or set attr if its not mode param
                setattr(obj, key, value)
        return obj

    def change_mode(self, mode_name: str, mode_value) -> bool:
        """Change mode of the Delivery object.
        Return True on success, False if `mode_name` was invalid.
        """
        if mode_name == 'email':
            self.__setattr__(mode_name, self.Email.from_dict(mode_value))
        elif mode_name == 'file':
            self.__setattr__(mode_name, self.File.from_dict(mode_value))
        elif mode_name == 'mobile':
            self.__setattr__(mode_name, self.Mobile.from_dict(mode_value))
        elif mode_name == 'ftp':
            self.__setattr__(mode_name, self.Ftp.from_dict(mode_value))
        elif mode_name == 'cache':
            self.__setattr__(mode_name, self.Cache.from_dict(mode_value))
        elif mode_name == 'history_list':
            self.__setattr__(mode_name, self.HistoryList.from_dict(mode_value))
        else:
            return False
        return True

    def _expiration_check(self, date):
        now = datetime.now().date()
        date = datetime.strptime(date, '%Y-%m-%d').date()

        return date >= now



# flake8: noqa
from .base_subscription import Subscription
from .cache_update_subscription import CacheUpdateSubscription
from .content import Content
from .delivery import (CacheType, ClientType, Delivery, Orientation, SendContentAs,
                       ShortcutCacheFormat, ZipSettings)
from .email_subscription import EmailSubscription
from .subscription_manager import list_subscriptions, SubscriptionManager



"""This is the demo script to show how administrator can manage subscriptions
and schedules.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import Connection
from mstrio.distribution_services import (Subscription, EmailSubscription, Content,
                                          SubscriptionManager, CacheUpdateSubscription, CacheType,
                                          list_subscriptions)
from mstrio.distribution_services import Schedule, list_schedules
from mstrio.distribution_services.subscription.cache_update_subscription import CacheUpdateSubscription

base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
conn = Connection(base_url, username, password, login_mode=1)

# create manager for subscriptions on a chosen project
sub_mngr = SubscriptionManager(connection=conn, project_name='MicroStrategy Tutorial')
# get all subscriptions from the given project (it is possible in two ways)
all_subs = list_subscriptions(connection=conn, project_name='MicroStrategy Tutorial')
all_subs = sub_mngr.list_subscriptions()

#  execute/delete subscriptions by passing theirs ids or Subscription objects
sub_mngr.execute(['11223344556677889900AABBCCDDEEFF', 'FFEEDDCCBBAA00998877665544332211'])
sub_mngr.delete(['11223344556677889900AABBCCDDEEFF', 'FFEEDDCCBBAA00998877665544332211'])

# list available recipients of the subscription for the given content (default
# delivery type is an email)
sub_mngr.available_recipients(content_id='11112222333344445555666677778888',
                              content_type='DOCUMENT')

# get a single subscription
sub = Subscription(connection=conn, subscription_id='AA11BB22CC33DD44EE55FF6677889900',
                   project_id='00FF99EE88DD77CC66BB55AA44332211')
# list all recipients of the given subscription and all available for this
# subscription
sub.recipients
sub.available_recipients()

# add/remove recipient(s) with given id(s)
sub.add_recipient(
    recipients=['1234567890A1234567890A1234567890', '98765432198765432198765432198765'])
sub.remove_recipient(
    recipients=['1234567890A1234567890A1234567890', '98765432198765432198765432198765'])

# execute a given subscription
sub.execute()

# replace a user with an admin in all of its subscriptions (e.g. when user exits
# company)
for s in sub_mngr.list_subscriptions(to_dictionary=False):
    if '9871239871298712413241235643132A' in [r['id'] for r in s.recipients]:
        s.add_recipient(recipients='11111111111111111111111111111111')
        s.remove_recipient(recipients='9871239871298712413241235643132A')

# create an email subscription
EmailSubscription.create(
    connection=conn, name="New Email Subscription for a Report",
    project_name="MicroStrategy Tutorial",
    contents=Content(id='ABC123ABC123ABC123ABC123ABC12345', type=Content.Type.REPORT,
                     personalization=Content.Properties(format_type='EXCEL')),
    schedules=['ABC123ABC123ABC123ABC123ABC12345'],
    recipients=['ABC123ABC123ABC123ABC123ABC12345'])

# create a cache update subscription
cache_update_sub = CacheUpdateSubscription(
    connection=conn, project_name='"MicroStrategy Tutorial',
    name=f'Cache Update Subscription for a Report', contents=Content(
        type=Content.Type.REPORT,
        id='ABC123ABC123ABC123ABC123ABC12345',
        personalization=Content.Properties(format_type=Content.Properties.FormatType.EXCEL),
    ), schedules=['ABC123ABC123ABC123ABC123ABC12345'], delivery_expiration_date='2025-12-31',
    send_now=True, recipients=['ABC123ABC123ABC123ABC123ABC12345'],
    cache_cache_type=CacheType.RESERVED)

# change name and owner of cache update subscription
cache_update_sub.alter(name=f"{cache_update_sub.name} (Altered)",
                       owner_id='AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA')

# list all cache update subscriptions
cache_update_subs = [
    sub for sub in list_subscriptions(conn, project_name="MicroStrategy Tutorial")
    if isinstance(sub, CacheUpdateSubscription)
]

# get list of schedules (you can filter them by for example name, id or
# description)
all_schdls = list_schedules(conn)

# get a single schedule by its id or name and then its properties
schdl = Schedule(connection=conn, name='Some shedule which runs daily')
schdl.list_properties()



from typing import Optional

from mstrio.utils.helper import Dictable


class Node(Dictable):

    def __init__(self, name: str, address: Optional[str] = None,
                 service_control: Optional[bool] = None) -> None:
        self.name = name
        self.address = address
        self.service_control = service_control



# flake8: noqa
from mstrio.server.project import *
from mstrio.utils import helper

helper.deprecation_warning(
    '`mstrio.server.application`',
    '`mstrio.server.project`',
    '11.3.4.101',  # NOSONAR
    False)



# flake8: noqa
from .project import (ProjectStatus, IdleMode, compare_project_settings, Application, Project,
                      ProjectSettings, ApplicationSettings, compare_application_settings)
from .cluster import GroupBy, ServiceAction, Cluster
from .environment import Environment
from .node import Node
from .server import ServerSettings
from .job_monitor import (Job, JobStatus, JobType, kill_all_jobs, kill_jobs, list_jobs, ObjectType,
                          PUName, SubscriptionType)



from enum import Enum
from pprint import pformat, pprint
from typing import Any, Dict, List, Optional, TypeVar, Union

from mstrio import config
from mstrio.api import subscriptions
from mstrio.connection import Connection
from mstrio.distribution_services.schedule import Schedule
from mstrio.distribution_services.subscription.content import Content
from mstrio.distribution_services.subscription.delivery import (CacheType, ClientType, Delivery,
                                                                LibraryCacheTypes, Orientation,
                                                                SendContentAs, ShortcutCacheFormat,
                                                                ZipSettings)
from mstrio.server.project import Project
from mstrio.users_and_groups import User
from mstrio.utils import helper, time_helper
from mstrio.utils.entity import EntityBase


class RecipientsTypes(Enum):
    CONTACT_GROUP = "CONTACT_GROUP"
    USER_GROUP = "USER_GROUP"
    CONTACT = "CONTACT"
    USER = "USER"
    PERSONAL_ADDRESS = "PERSONAL_ADDRESS"
    UNSUPPORTED = "UNSUPPORTED"


T = TypeVar("T")


class Subscription(EntityBase):
    """Class representation of MicroStrategy Subscription object.

    Attributes:
        id: The ID of the Subscription
        connection: The MicroStrategy connection object
        project_id: The ID of the project the Subscription belongs to
    """

    _API_GETTERS = {
        ("id", "name", "editable", "date_created", "date_modified", "owner", "schedules",
         "contents", "recipients", "delivery"): subscriptions.get_subscription
    }
    _FROM_DICT_MAP = {
        "owner": User.from_dict,
        "contents": lambda source, connection:
            [Content.from_dict(content, connection) for content in source],
        "delivery": Delivery.from_dict,
        "schedules": lambda source, connection:
            [Schedule.from_dict(content, connection) for content in source],
        "date_created": time_helper.DatetimeFormats.YMDHMS,
        "date_modified": time_helper.DatetimeFormats.YMDHMS,
    }
    _API_PATCH = [subscriptions.update_subscription]
    _RECIPIENTS_TYPES = RecipientsTypes
    _RECIPIENTS_INCLUDE = ['TO', 'CC', 'BCC', None]

    def __init__(self, connection: Connection, subscription_id: str, project_id: str = None,
                 project_name: str = None, application_id: str = None,
                 application_name: str = None):
        """Initialize Subscription object, populates it with I-Server data.
        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name`
        is omitted.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
            subscription_id: ID of the subscription to be initialized
            project_id: Project ID
            project_name: Project name
            application_id: deprecated. Use project_id instead.
            application_name: deprecated. Use project_name instead.
        """
        if application_id or application_name:
            helper.deprecation_warning(
                '`application`',
                '`project`',
                '11.3.4.101',  # NOSONAR
                False)
            project_id = project_id or application_id
            project_name = project_name or application_name

        project_id = self._project_id_check(connection, project_id, project_name)
        super().__init__(connection, subscription_id, project_id=project_id)

    def _init_variables(self, project_id, **kwargs):
        super()._init_variables(**kwargs)
        self.subscription_id = kwargs.get('id')
        self.editable = kwargs.get('editable')
        self.allow_delivery_changes = kwargs.get('allow_delivery_changes')
        self.allow_personalization_changes = kwargs.get('allow_personalization_changes')
        self.allow_unsubscribe = kwargs.get('allow_unsubscribe')
        self.date_created = time_helper.map_str_to_datetime("date_created",
                                                            kwargs.get("date_created"),
                                                            self._FROM_DICT_MAP)
        self.date_modified = time_helper.map_str_to_datetime("date_modified",
                                                             kwargs.get("date_modified"),
                                                             self._FROM_DICT_MAP)
        self.owner = User.from_dict(kwargs.get('owner'),
                                    self.connection) if kwargs.get('owner') else None
        self.schedules = [
            Schedule.from_dict(schedule, self._connection) for schedule in kwargs.get('schedules')
        ] if kwargs.get('schedules') else None
        self.contents = [
            Content.from_dict(content, self._connection) for content in kwargs.get('contents')
        ] if kwargs.get('contents') else None
        self.recipients = kwargs.get('recipients', None)
        self.delivery = Delivery.from_dict(
            kwargs.get('delivery')) if kwargs.get('delivery') else None
        self.project_id = project_id

    def alter(
        self,
        name: Optional[str] = None,
        allow_delivery_changes: Optional[bool] = None,
        allow_personalization_changes: Optional[bool] = None,
        allow_unsubscribe: Optional[bool] = None,
        send_now: bool = False,
        owner_id: Optional[str] = None,
        schedules: Optional[Union[str, List[str], Schedule, List[Schedule]]] = None,
        contents: Optional[Content] = None,
        recipients: Optional[Union[List[str], List[dict]]] = None,
        delivery: Optional[Union[Delivery, dict]] = None,
        delivery_mode: Optional[str] = None,
        custom_msg: Optional[str] = None,
        delivery_expiration_date: Optional[str] = None,
        contact_security: Optional[bool] = None,
        filename: Optional[str] = None,
        compress: Optional[bool] = None,
        space_delimiter: Optional[str] = None,
        email_subject: Optional[str] = None,
        email_message: Optional[str] = None,
        email_send_content_as: Optional[str] = None,
        overwrite_older_version: Optional[bool] = None,
        zip_filename: Optional[str] = None,
        zip_password_protect: Optional[bool] = None,
        zip_password: Optional[str] = None,
        file_burst_sub_folder: Optional[str] = None,
        printer_copies: Optional[int] = None,
        printer_range_start: Optional[int] = None,
        printer_range_end: Optional[int] = None,
        printer_collated: Optional[bool] = None,
        printer_orientation: Optional[str] = None,
        printer_use_print_range: Optional[bool] = None,
        cache_cache_type: Optional[str] = None,
        cache_shortcut_cache_format: Optional[str] = None,
        mobile_client_type: Optional[str] = None,
        device_id: Optional[str] = None,
        do_not_create_update_caches: Optional[bool] = None,
        re_run_hl: Optional[bool] = None,
        cache_library_cache_types: List[Union[LibraryCacheTypes, str]] = [LibraryCacheTypes.WEB],
        cache_reuse_dataset_cache: bool = False,
        cache_is_all_library_users: bool = False,
        delivery_notification_enabled: bool = False,
        delivery_personal_notification_address_id: Optional[str] = None,
    ):
        """
        Alter subscription.

        Args:
            connection(Connection): a MicroStrategy connection object
            name(str): name of the subscription,
            project_id(str): project ID,
            allow_delivery_changes(bool): whether the recipients can change
                the delivery of the subscription,
            allow_personalization_changes(bool): whether the recipients can
                personalize the subscription,
            allow_unsubscribe(bool): whether the recipients can unsubscribe
                from the subscription,
            send_now(bool): indicates whether to execute the subscription
                immediately,
            owner_id(str): ID of the subscription owner, by default logged in
                user ID,
            schedules (Union[str, List[str], Schedule, List[Schedule]]):
                Schedules IDs or Schedule objects,
            contents (Content): The content of the subscription.
            recipients (Union[List[str], List[dict]]): list of recipients IDs
                or dicts,
            delivery_mode(str, enum): the subscription delivery mode [EMAIL,
                FILE, PRINTER, HISTORY_LIST, CACHE, MOBILE, FTP, SNAPSHOT,
                PERSONAL_VIEW, SHARED_LINK, UNSUPPORTED],
            delivery_expiration_date(str): expiration date of the subscription,
                format should be yyyy-MM-dd,
            contact_security(bool): whether to use contact security for each
                contact group member,
            filename(str): the filename that will be delivered when
                the subscription is executed,
            compress(bool): whether to compress the file
            space_delimiter(str): space delimiter,
            email_subject(str): email subject associated with the subscription,
            email_message(str): email body of subscription,
            email_send_content_as(str,enum): [data, data_and_history_list,
                data_and_link_and_history_list, link_and_history_list],
            overwrite_older_version(bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            zip_filename(str): filename of the compressed content,
            zip_password_protect(bool): whether to password protect zip file,
            zip_password(str): optional password for the compressed file
            file_burst_sub_folder(str): burst sub folder,
            printer_copies(int): the number of copies that should be printed,
            printer_range_start(int): the number indicating the first report
                page that should be printed,
            printer_range_end(int): the number indicating the last report
                page that should be printed,
            printer_collated(bool): whether the printing should be collated,
            printer_orientation(str,enum): [ PORTRAIT, LANDSCAPE ]
            printer_use_print_range(bool): whether print range should be used,
            cache_cache_type(str,enum): [RESERVED, SHORTCUT,
                SHORTCUTWITHBOOKMARK]
            cache_shortcut_cache_format(str,enum): [RESERVED, JSON, BINARY,
                BOTH]
            mobile_client_type(str,enum): [RESERVED, BLACKBERRY, PHONE, TABLET,
                ANDROID]
            device_id(str): the mobile target project,
            do_not_create_update_caches(bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            re_run_hl(bool): whether subscription will re-run against warehouse
            cache_library_cache_types: Set of library cache types,
                available types can be web, android, ios
            cache_reuse_dataset_cache: Whether to reuse dataset cache
            cache_is_all_library_users: Whether for all library users
            delivery_notification_enabled: Whether notification is enabled,
                notification applies to cache
            delivery_personal_notification_address_id: Notification details
        """

        def is_changed(nested=None, **kwargs):
            for key, value in kwargs.items():
                if nested:
                    return value if value != nested and value is not None else nested
                else:
                    current_val = self.__dict__.get(key)
                    # if not current_val: we need to get
                    return value if value != current_val and value is not None else current_val

        # Schedules logic
        schedules = self.__validate_schedules(schedules=schedules)
        if not schedules:
            schedules = [{'id': sch.id} for sch in self.schedules]

        # Content logic
        if contents:
            contents = contents if isinstance(contents, list) else [contents]
            content_type_msg = "Contents must be dictionaries or Content objects."
            contents = [
                content.to_dict(
                    camel_case=True) if isinstance(content, Content) else content if isinstance(
                        content, dict) else helper.exception_handler(content_type_msg, TypeError)
                for content in contents
            ]
        else:
            contents = [cont.to_dict() for cont in self.contents]

        # Delivery logic
        if delivery:
            temp_delivery = (Delivery.from_dict(delivery)
                             if isinstance(delivery, dict) else delivery)
        else:
            temp_delivery = self.__change_delivery_properties(
                delivery_mode, delivery_expiration_date, contact_security, email_subject,
                email_message, filename, compress, None, zip_filename, zip_password,
                zip_password_protect, space_delimiter, email_send_content_as,
                overwrite_older_version, file_burst_sub_folder, printer_copies,
                printer_range_start, printer_range_end, printer_collated, printer_orientation,
                printer_use_print_range, client_type=mobile_client_type, device_id=device_id,
                do_not_create_update_caches=do_not_create_update_caches, re_run_hl=re_run_hl,
                cache_type=cache_cache_type, shortcut_cache_format=cache_shortcut_cache_format,
                library_cache_types=cache_library_cache_types,
                reuse_dataset_cache=cache_reuse_dataset_cache,
                is_all_library_users=cache_is_all_library_users,
                notification_enabled=delivery_notification_enabled,
                personal_notification_address_id=delivery_personal_notification_address_id)
        delivery = temp_delivery.to_dict(camel_case=True)

        # Recipients logic
        recipients = is_changed(recipients=recipients)
        recipients = Subscription._validate_recipients(self.connection, contents, recipients,
                                                       self.project_id, delivery['mode'])

        body = {
            "name": is_changed(name=name),
            "allowDeliveryChanges": is_changed(allow_delivery_changes=allow_delivery_changes),
            "allowPersonalizationChanges":
                is_changed(allow_personalization_changes=allow_personalization_changes),
            "allowUnsubscribe": is_changed(allow_unsubscribe=allow_unsubscribe),
            "sendNow": send_now,
            'owner': {
                'id': is_changed(nested=self.owner.id, owner_id=owner_id)
            },
            "schedules": schedules,
            "contents": contents,
            "recipients": recipients,
            "delivery": delivery,
        }

        body = helper.delete_none_values(body)

        response = subscriptions.update_subscription(self.connection, self.id, self.project_id,
                                                     body)

        if response.ok:
            response = response.json()
            response = helper.camel_to_snake(response)
            self._set_object(**response)
            if config.verbose:
                print(custom_msg if custom_msg else "Updated subscription '{}' with ID: {}."
                      .format(self.name, self.id))

    @staticmethod
    def __validate_schedules(schedules: Union[str, List[str], Schedule, List[Schedule]] = None):
        tmp_schedules = []
        schedules = schedules if isinstance(schedules, list) else [schedules]
        schedules = [s for s in schedules if s is not None]
        for schedule in schedules:
            if isinstance(schedule, Schedule):
                sch_id = schedule.id
            elif isinstance(schedule, str):
                sch_id = schedule
            tmp_schedules.append({'id': sch_id})

        return tmp_schedules

    def list_properties(self):
        """Lists all properties of subscription."""
        return {
            key: self.__dict__[key]
            for key in sorted(self.__dict__, key=helper.sort_object_properties)
            if key not in ['connection', 'application_id', 'project_id', '_delivery']
        }

    def execute(self):
        """Executes a subscription with given name or GUID for given project.
        """
        self.alter(
            send_now=True,
            custom_msg="Executed subscription '{}' with ID '{}'.".format(self.name, self.id))

    def delete(self, force: bool = False) -> bool:
        """Delete a subscription. Returns True if deletion was successful.

        Args:
            force: If True, no additional prompt will be shown before deleting
        """
        user_input = 'N'
        if not force:
            user_input = input(
                "Are you sure you want to delete subscription '{}' with ID: {}? [Y/N]: ".format(
                    self.name, self.id))
        if force or user_input == 'Y':
            response = subscriptions.remove_subscription(self.connection, self.id, self.project_id)
            if response.ok and config.verbose:
                print("Deleted subscription '{}' with ID: {}.".format(self.name, self.id))
            return response.ok
        else:
            return False

    def available_bursting_attributes(self) -> List[str]:
        """Get a list of available attributes for bursting feature."""
        contents_bursting = {}
        for content in self.contents:
            response = subscriptions.bursting_attributes(
                self.connection,
                self.project_id,
                content.id,
                content.type.upper(),
            )
            if response.ok:
                contents_bursting[content.id] = response.json()['burstingAttributes']

        return contents_bursting

    def available_recipients(self) -> List[dict]:
        """List available recipients for subscription content."""
        body = {"contents": [content.to_dict() for content in self.contents]}
        delivery_type = self.delivery.mode
        response = subscriptions.available_recipients(self.connection, self.project_id, body,
                                                      delivery_type)

        if response.ok and config.verbose:
            return response.json()['recipients']

    def add_recipient(self, recipients: Union[List[dict], dict, List[str],
                                              str] = [], recipient_id: Optional[str] = None,
                      recipient_type: Optional[str] = None, recipient_include_type: str = 'TO'):
        """Adds recipient to subscription. You can either specify id, type and
        include_type of single recipient, or just pass recipients list as a
        list of dictionaries.

        Args:
            recipients: list of ids or dicts containing recipients, dict format:
                {"id": recipient_id,
                 "type": "CONTACT_GROUP" / "USER_GROUP" / "CONTACT" /
                         "USER" / "PERSONAL_ADDRESS" / "UNSUPPORTED"
                 "includeType": "TO" / "CC" / "BCC"}
            recipient_id: id of the recipient
            recipient_type: type of the recipient
            recipient_include_type: include type of the recipient one of the
                following: "TO" / "CC" / "BCC", by default "TO"
        """
        if len(recipients) > 0:
            recipients = recipients if isinstance(recipients, list) else [recipients]
        if recipient_id and recipient_type:
            recipients.append({
                "id": recipient_id,
                "type": recipient_type,
                "includeType": recipient_include_type
            })
        elif recipients == [] and recipient_id is None:
            msg = ("Specify either a recipient ID, type and include type or pass recipients "
                   "dictionaries")
            helper.exception_handler(msg, ValueError)
        elif len(recipients) >= 1 and recipient_id:
            msg = ("Specify either a recipient ID, type and include type or pass recipients "
                   "dictionaries")
            helper.exception_handler(msg, ValueError)

        all_recipients = self.recipients.copy()
        ready_recipients = []
        exisiting_recipients = [rec['id'] for rec in self.recipients]
        for recipient in recipients:
            not_dict_msg = """Wrong recipient format, expected format is
                              {"id": recipient_id,
                               "type": "CONTACT_GROUP" / "USER_GROUP" / "CONTACT" /
                                       "USER" / "PERSONAL_ADDRESS" / "UNSUPPORTED"
                               "includeType": "TO" / "CC" / "BCC" (optional)
                              }"""
            if isinstance(recipient, dict):
                if list(recipient.keys()) in [['id', 'type', 'includeType'], ['id', 'type']]:
                    rec_id = recipient['id']
                    if rec_id in exisiting_recipients:
                        helper.exception_handler(
                            "{} is already a recipient of subscription".format(rec_id),
                            UserWarning)
                    else:
                        ready_recipients.append(recipient)
                else:
                    helper.exception_handler(not_dict_msg, TypeError)
            if isinstance(recipient, str):
                if recipient in exisiting_recipients:
                    helper.exception_handler(
                        "{} is already a recipient of subscription".format(recipient), UserWarning)
                else:
                    ready_recipients.append(recipient)

        ready_recipients = self._validate_recipients(self.connection, self.contents,
                                                     ready_recipients, self.project_id,
                                                     self.delivery.mode)

        if ready_recipients:
            all_recipients.extend(ready_recipients)
            self.alter(recipients=all_recipients)
        elif config.verbose:
            print("No recipients were added to the subscription.")

    def remove_recipient(self, recipients):
        """Removes recipient from given subscription in given project.

        Args:
            recipients: list of ids or dicts containing recipients, dict format:
                {"id": recipient_id,
                 "type": "CONTACT_GROUP" / "USER_GROUP" / "CONTACT" /
                         "USER" / "PERSONAL_ADDRESS" / "UNSUPPORTED"
                 "includeType": "TO" / "CC" / "BCC"}
        """
        all_recipients = self.recipients
        recipients = recipients if isinstance(recipients, list) else [recipients]
        exisiting_recipients = [rec['id'] for rec in self.recipients]

        if len(self.recipients) == 1:
            helper.exception_handler(
                "Subscription must have at last one recipient. Add new recipient before removing.")
        for recipient in recipients:
            rec_id = recipient['id'] if isinstance(recipient, dict) else recipient
            if rec_id not in exisiting_recipients:
                helper.exception_handler("{} is not a recipient of subscription".format(rec_id),
                                         UserWarning)
            else:
                all_recipients = [rec for rec in all_recipients if rec['id'] != rec_id]
        if len(all_recipients) == 0:
            helper.exception_handler(
                "You cannot remove all existing recipients. Add new recipient before removing.")
        elif len(self.recipients) - len(all_recipients) > 0:
            self.alter(recipients=all_recipients)
        elif len(self.recipients) == len(all_recipients) and config.verbose:
            print("No recipients were removed from the subscription.")

    def __change_delivery_properties(
        self,
        mode=None,
        expiration=None,
        contact_security=None,
        subject: Optional[str] = None,
        message: Optional[str] = None,
        filename: Optional[str] = None,
        compress: Optional[bool] = None,
        zip_settings: Optional[ZipSettings] = None,
        zip_filename: Optional[str] = None,
        zip_password: Optional[str] = None,
        zip_password_protect: Optional[bool] = None,
        space_delimiter: Optional[str] = None,
        send_content_as: Optional[SendContentAs] = None,
        overwrite_older_version: Optional[bool] = None,
        burst_sub_folder: Optional[str] = None,
        copies: Optional[int] = None,
        range_start: Optional[int] = None,
        range_end: Optional[int] = None,
        collated: Optional[bool] = None,
        orientation: Optional[Orientation] = None,
        use_print_range: Optional[bool] = None,
        cache_type: Optional[CacheType] = None,
        shortcut_cache_format: Optional[ShortcutCacheFormat] = None,
        client_type: Optional[ClientType] = None,
        device_id: Optional[str] = None,
        do_not_create_update_caches: Optional[bool] = None,
        re_run_hl: Optional[bool] = None,
        library_cache_types: List[Union[LibraryCacheTypes, str]] = [LibraryCacheTypes.WEB],
        reuse_dataset_cache: bool = False,
        is_all_library_users: bool = False,
        notification_enabled: bool = False,
        personal_notification_address_id: Optional[str] = None,
    ):

        func = self.__change_delivery_properties
        args = func.__code__.co_varnames[:func.__code__.co_argcount]
        defaults = func.__defaults__
        default_dict = dict(zip(args[-len(defaults):], defaults)) if defaults else {}
        local = locals()
        # create dict of properties to be changed
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]

        # 'zip_settings' is 'zip' in object
        if 'zip_settings' in properties.keys():
            properties['zip'] = properties.pop('zip_settings')

        obj_dict = self.delivery.VALIDATION_DICT
        obj_mode_dict = self.delivery.__dict__[self.delivery.mode.lower()].VALIDATION_DICT
        obj_mode_zip_dict = {
            "zip_filename": str,
            "zip_password": str,
            "zip_password_protect": bool,
        }

        if properties:  # at this point only not None values in properties
            for key, value in properties.items():
                if key in obj_dict.keys():  # Highest level parameters. Mainly mode type and object
                    # Change mode or set attr if it's not mode param
                    if not self.delivery.change_mode(key, value):
                        self.delivery.__setattr__(key, value)
                elif key in obj_mode_dict.keys():  # parameters of a particular mode e.g. of email
                    helper.rsetattr(self.delivery, f'{self.delivery.mode.lower()}.{key}', value)
                elif key in obj_mode_zip_dict.keys():  # zip settings
                    key = key[4:]
                    if not helper.rgetattr(self.delivery, f'{self.delivery.mode.lower()}.zip',
                                           None):
                        helper.rsetattr(self.delivery, f'{self.delivery.mode.lower()}.zip',
                                        ZipSettings())
                    helper.rsetattr(self.delivery, f'{self.delivery.mode.lower()}.zip.{key}',
                                    value)
        return self.delivery

    @classmethod
    def from_dict(cls: T, source: Dict[str, Any] = None, connection: Optional["Connection"] = None,
                  project_id: Optional[str] = None, project_name: Optional[str] = None,
                  dictionary=None, application_id: Optional[str] = None,
                  application_name: Optional[str] = None) -> T:
        """Initialize Subscription object from dictionary.
        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is
        omitted."""
        if application_id or application_name:
            helper.deprecation_warning(
                '`application_id` and `application_name`',
                '`project_id` and `project_name`',
                '11.3.4.101',  # NOSONAR
                False)
            project_id = project_id or application_id
            project_name = project_name or application_name

        if dictionary:
            helper.deprecation_warning(
                "`dictionary`",
                "`source`",
                "11.3.4.101",  # NOSONAR
                module=False,
            )
        source = source or dictionary
        # This is tricky, as we have to consider 3 cases:
        # - project_id given directly
        # - project_name given in args
        # - project_id in dict (if someone serialised to dict
        #                           and is now deserialising)
        try:
            project_id = Subscription._project_id_check(connection, project_id, project_name)
        except ValueError as err:
            if source.get("project_id", False):
                project_id = source["project_id"]
            else:
                raise err
        _source = {
            **source,
            "project_id": project_id,
        }
        obj: Subscription = super().from_dict(_source, connection)

        return obj

    @classmethod
    def __create(
        cls,
        connection: Connection,
        name: str,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        allow_delivery_changes: Optional[bool] = None,
        allow_personalization_changes: Optional[bool] = None,
        allow_unsubscribe: Optional[bool] = None,
        send_now: Optional[bool] = None,
        owner_id: Optional[str] = None,
        schedules: Union[str, List[str], Schedule, List[Schedule]] = None,
        contents: Content = None,
        recipients: Union[List[dict], List[str]] = None,
        delivery: Union[Delivery, dict] = None,
        delivery_mode: str = Delivery.DeliveryMode.EMAIL,
        delivery_expiration_date: Optional[str] = None,
        contact_security: bool = True,
        filename: Optional[str] = None,
        compress: bool = False,
        space_delimiter: Optional[str] = None,
        email_subject: Optional[str] = None,
        email_message: Optional[str] = None,
        email_send_content_as: str = SendContentAs.DATA,
        overwrite_older_version: bool = False,
        zip_filename: Optional[str] = None,
        zip_password_protect: Optional[bool] = None,
        zip_password: Optional[str] = None,
        file_burst_sub_folder: Optional[str] = None,
        printer_copies: int = 1,
        printer_range_start: int = 0,
        printer_range_end: int = 0,
        printer_collated: bool = True,
        printer_orientation: str = Orientation.PORTRAIT,
        printer_use_print_range: bool = False,
        cache_cache_type: Union[CacheType, str] = CacheType.RESERVED,
        cache_shortcut_cache_format: Union[ShortcutCacheFormat,
                                           str] = ShortcutCacheFormat.RESERVED,
        mobile_client_type: str = ClientType.RESERVED,
        device_id: Optional[str] = None,
        do_not_create_update_caches: bool = True,
        re_run_hl: bool = True,
        cache_library_cache_types: List[Union[LibraryCacheTypes, str]] = [LibraryCacheTypes.WEB],
        cache_reuse_dataset_cache: bool = False,
        cache_is_all_library_users: bool = False,
        delivery_notification_enabled: bool = False,
        delivery_personal_notification_address_id: Optional[str] = None,
    ):
        """Creates a subscription Create_Subscription_Outline.

        Args:
            connection(Connection): a MicroStrategy connection object
            name(str): name of the subscription,
            project_id(str): project ID,
            project_name(str): project name,
            allow_delivery_changes(bool): whether the recipients can change
                the delivery of the subscription,
            allow_personalization_changes(bool): whether the recipients can
                personalize the subscription,
            allow_unsubscribe(bool): whether the recipients can unsubscribe
                from the subscription,
            send_now(bool): indicates whether to execute the subscription
                immediately,
            owner_id(str): ID of the subscription owner, by default logged in
                user ID,
            schedules (Union[str, List[str], Schedule, List[Schedule]]):
                Schedules IDs or Schedule objects,
            contents (Content): The content settings.
            recipients (List[dict],List[str]): list of recipients IDs or dicts,
            delivery(Union[Delivery,dict]): delivery object or dict
            delivery_mode(str, enum): the subscription delivery mode [EMAIL,
                FILE, PRINTER, HISTORY_LIST, CACHE, MOBILE, FTP, SNAPSHOT,
                PERSONAL_VIEW, SHARED_LINK, UNSUPPORTED],
            delivery_expiration_date(str): expiration date of the subscription,
                format should be yyyy-MM-dd,
            contact_security(bool): whether to use contact security for each
                contact group member,
            filename(str): the filename that will be delivered when
                the subscription is executed,
            compress(bool): whether to compress the file,
            space_delimiter(str): space delimiter,
            email_subject(str): email subject associated with the subscription,
            email_message(str): email body of subscription,
            email_send_content_as(str,enum): [data, data_and_history_list,
                data_and_link_and_history_list, link_and_history_list],
            overwrite_older_version(bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            zip_filename(str): filename of the compressed content,
            zip_password_protect(bool): whether to password protect zip file,
            zip_password(str): optional password for the compressed file,
            file_burst_sub_folder(str): burst sub folder,
            printer_copies(int): the number of copies that should be printed,
            printer_range_start(int): the number indicating the first report
                page that should be printed,
            printer_range_end(int): the number indicating the last report
                page that should be printed,
            printer_collated(bool): whether the printing should be collated,
            printer_orientation(str,enum): [ PORTRAIT, LANDSCAPE ]
            printer_use_print_range(bool): whether print range should be used,
            cache_cache_type(str,enum): [RESERVED, SHORTCUT,
                SHORTCUTWITHBOOKMARK]
            cache_shortcut_cache_format(str,enum):[RESERVED, JSON, BINARY, BOTH]
            mobile_client_type(str,enum): [RESERVED, BLACKBERRY, PHONE, TABLET,
                ANDROID]
            device_id(str): the mobile target project,
            do_not_create_update_caches(bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            re_run_hl(bool): whether subscription will re-run against warehouse
            cache_library_cache_types: Set of library cache types,
                available types can be web, android, ios
            cache_reuse_dataset_cache: Whether to reuse dataset cache
            cache_is_all_library_users: Whether for all library users
            delivery_notification_enabled: Whether notification is enabled,
                notification applies to cache
            delivery_personal_notification_address_id: Notification details
        """
        name = name if len(name) <= 255 else helper.exception_handler(
            "Name too long. Max name length is 255 characters.")
        project_id = cls._project_id_check(connection, project_id, project_name)

        if not schedules:
            msg = ("Please specify 'schedules' parameter.")
            helper.exception_handler(msg)

        schedules = cls.__validate_schedules(schedules=schedules)

        # Content logic
        contents = contents if isinstance(contents, list) else [contents]
        content_type_msg = "Contents must be dictionaries or Content objects."
        contents = [
            content.to_dict(camel_case=True) if isinstance(content, Content) else content if
            isinstance(content, dict) else helper.exception_handler(content_type_msg, TypeError)
            for content in contents
        ]

        # Delivery logic
        if delivery:
            temp_delivery = Delivery.from_dict(delivery) if isinstance(delivery,
                                                                       dict) else delivery
        else:
            temp_delivery = Delivery(
                delivery_mode, delivery_expiration_date, contact_security, email_subject,
                email_message, filename, compress, None, zip_password, zip_password_protect,
                space_delimiter, email_send_content_as, overwrite_older_version,
                file_burst_sub_folder, printer_copies, printer_range_start, printer_range_end,
                printer_collated, printer_orientation, printer_use_print_range, mobile_client_type,
                device_id, do_not_create_update_caches, re_run_hl, cache_type=cache_cache_type,
                shortcut_cache_format=cache_shortcut_cache_format,
                library_cache_types=cache_library_cache_types,
                reuse_dataset_cache=cache_reuse_dataset_cache,
                is_all_library_users=cache_is_all_library_users,
                notification_enabled=delivery_notification_enabled,
                personal_notification_address_id=delivery_personal_notification_address_id)
        delivery = temp_delivery.to_dict(camel_case=True)

        # Recipients logic
        recipients = cls._validate_recipients(connection, contents, recipients, project_id,
                                              delivery['mode'])

        # Create body
        body = {
            "name": name,
            "allowDeliveryChanges": allow_delivery_changes,
            "allowPersonalizationChanges": allow_personalization_changes,
            "allowUnsubscribe": allow_unsubscribe,
            "sendNow": send_now,
            "owner": {
                "id": owner_id
            },
            "schedules": schedules,
            "contents": contents,
            "recipients": recipients,
            "delivery": delivery
        }

        body = helper.delete_none_values(body)
        response = subscriptions.create_subscription(connection, project_id, body)
        if config.verbose:
            unpacked_response = response.json()
            print("Created subscription '{}' with ID: '{}'.".format(name, unpacked_response['id']))
        return cls.from_dict(response.json(), connection, project_id)

    @staticmethod
    def _validate_recipients(connection, contents: List[Union[Content, dict]], recipients,
                             project_id, delivery_mode):
        recipients = recipients if isinstance(recipients, list) else [recipients]
        body = {
            "contents": [
                cont.to_dict() if isinstance(cont, Content) else cont for cont in contents
            ]
        }
        available_recipients = subscriptions.available_recipients(connection, project_id, body,
                                                                  delivery_mode)
        available_recipients = available_recipients.json()['recipients']
        available_recipients_ids = [rec['id'] for rec in available_recipients]
        # Format recipients list if needed
        formatted_recipients = []
        if recipients:
            for recipient in recipients:
                if isinstance(recipient, dict):
                    if recipient['id'] in available_recipients_ids:
                        formatted_recipients.append(recipient)
                    else:
                        msg = (f"'{recipient['id']}' is not a valid recipient ID for selected "
                               "content and delivery mode. Available recipients: \n"
                               f"{pformat(available_recipients, indent=2)}")
                        helper.exception_handler(msg, ValueError)
                        pprint(available_recipients)
                elif isinstance(recipient, str):
                    if recipient in available_recipients_ids:
                        rec = helper.filter_list_of_dicts(available_recipients, id=recipient)
                        formatted_recipients.append(rec[0])
                    else:
                        msg = (f"'{recipient}' is not a valid recipient ID for selected content "
                               "and delivery mode. Available recipients: \n"
                               f"{pformat(available_recipients,indent=2)}")
                        helper.exception_handler(msg, ValueError)
                        pprint(available_recipients)
                else:
                    helper.exception_handler(
                        "Recipients must be a dictionaries or a strings, not a {}".format(
                            type(recipient)), exception_type=TypeError)
        return formatted_recipients

    @staticmethod
    def _project_id_check(connection, project_id, project_name):
        """Check if the project name exists and returns the project ID.

        Args:
            connection(object): MicroStrategy connection object
            project_id: Project ID
            project_name: Project name
        """
        if project_id is None and project_name is None:
            msg = ("Please specify either 'project_name' or 'project_id' "
                   "parameter in the constructor.")
            helper.exception_handler(msg)
        if project_id is None:
            project_loaded_list = Project._list_loaded_projects(connection, to_dictionary=True,
                                                                name=project_name)
            try:
                project_id = project_loaded_list[0]['id']
            except IndexError:
                helper.exception_handler(
                    "There is no project with the given name: '{}'".format(project_name),
                    exception_type=ValueError)

        return project_id



"""This is the demo script to show how administrator can manage server and the
cluster with its nodes.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import Connection
from mstrio.server import Cluster, Environment, Project

base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
conn = Connection(base_url, username, password, login_mode=1)
MICROSTRATEGY_TUTORIAL = 'MicroStrategy Tutorial'

# get the cluster for given connection
clstr = Cluster(connection=conn)

# save topologies of services or nodes from this cluster into a dataframe
nodes_topology_df = clstr.nodes_topology()
services_topology_df = clstr.services_topology()

# get list of services grouped by nodes or by services
services_by_nodes = clstr.list_services(group_by='nodes')
services_by_services = clstr.list_services(group_by='services')

# get list of nodes (information about projects within each node is given there)
nodes = clstr.list_nodes(to_dictionary=True)
# remove/add a node from/to a cluster (node with the given name should exist)
clstr.remove_node(name='env-xxxxxxlaio3use1')
clstr.add_node(name='env-xxxxxxlaio3use1')

# get name of default (primary) node of the cluster and set new deafult node
clstr.default_node
clstr.set_primary_node(name='env-xxxxxxlaio3use1')

# update node settings or reset them to default values
clstr.update_node_settings(node='env-xxxxxxlaio3use1', load_balance_factor=99,
                           initial_pool_size=511, max_pool_size=1023)
clstr.reset_node_settings(node='env-xxxxxxlaio3use1')

# stop/start service on selected nodes (error will be thrown in case of wrong
# names of service or nodes)
clstr.stop(service='Apache-Kafka', nodes=['env-xxxxxxlaio1use1', 'env-xxxxxxlaio2use1'])
clstr.start(service='Apache-Kafka', nodes=['env-xxxxxxlaio1use1', 'env-xxxxxxlaio2use1'])

env = Environment(connection=conn)
# list all projects available for the given connection (it is possible via
# class Cluster or Environment)
projects = env.list_projects()
projects = clstr.list_projects()

# load or unload chosen project (it is possible via class Cluster or
# Project)
project = Project(connection=conn, name=MICROSTRATEGY_TUTORIAL)
project.load()
project.unload()

# via Cluster can we also specify on which node(s) project will be loaded
# or unloaded
clstr.load_project(project=MICROSTRATEGY_TUTORIAL,
                   on_nodes=['env-xxxxxxlaio1use1', 'env-xxxxxxlaio2use1'])
clstr.unload_project(project=MICROSTRATEGY_TUTORIAL,
                     on_nodes=['env-xxxxxxlaio1use1', 'env-xxxxxxlaio2use1'])

# get settings of a server as a dataframe
server_settings_df = env.server_settings.to_dataframe

# save/load settings of a server to/from a file (format can be 'csv', 'json' or
# 'pickle')
env.server_settings.to_csv(name="server_settings.csv")
env.server_settings.import_from(file="server_settings.csv")

# update some settings of a server
env.server_settings.allowUserLoginWithFullName = True
env.server_settings.minCharsPasswordChanges = 1
env.server_settings.update()



"""This is the demo script to show how administrator can manage users and
user groups.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import Connection
from mstrio.users_and_groups import User, list_users, UserGroup, list_user_groups

base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
conn = Connection(base_url, username, password, project_name="MicroStrategy Tutorial",
                  login_mode=1)

# create multiple users
users_array = [{
    'username': 'jwilliams',
    'fullName': 'John Williams'
}, {
    'username': 'mjones',
    'fullName': 'Mark Jones'
}, {
    'username': 'sbrown',
    'fullName': 'Steve Brown'
}, {
    'username': 'jdavis',
    'fullName': 'James Davis'
}, {
    'username': 'twilson',
    'fullName': 'Thomas Wilson'
}]
for u in users_array:
    User.create(connection=conn, username=u['username'], full_name=u['fullName'])

# create a single user and get users which name begins with "John" and have
# additional filter for initials
User.create(connection=conn, username="jsmith", full_name="John Smith")
my_users = list_users(connection=conn, name_begins="John", initials="JS")

# get all user groups (you can also add additional filters as for users) and
# create a new one
user_groups_list = list_user_groups(connection=conn)
UserGroup.create(connection=conn, name="Special Users")

# get user, user group and add this user to this user group
user_ = User(connection=conn, name="John Smith")
user_group_ = UserGroup(connection=conn, name="Special Users")
user_group_.add_users(users=[user_.id])

# set custom permissions of the user for given objects
user_.set_custom_permissions(to_objects=['55AA293811EAE2F2EC7D0080EF25A592'], object_type=3,
                             execute='grant', use='deny', control='default', delete='grant',
                             write='deny', read='default', browse='grant')

# set permission of the user group for given objects
user_group_.set_permission(permission='Full Control',
                           to_objects=['55AA293811EAE2F2EC7D0080EF25A592'], object_type=3)

# delete a user and a user group
user_.delete()
user_group_.delete()



from typing import Any, Callable, Dict, List, Optional, Union

from requests import HTTPError

from mstrio import config
from mstrio.api import datasources, objects
from mstrio.connection import Connection
from mstrio.datasources import DatasourceConnection, DatasourceInstance, DatasourceLogin
from mstrio.server.project import Project
from mstrio.users_and_groups.user import User
from mstrio.utils import helper
from mstrio.utils.entity import Entity, EntityBase, ObjectTypes
from mstrio.utils.helper import get_objects_id


def list_datasource_mappings(
        connection: Connection, default_connection_map: bool = False,
        project: Optional[Union[Project, str]] = None, to_dictionary: bool = False,
        limit: Optional[int] = None, application: Optional[Union[Project, str]] = None,
        **filters) -> Union["DatasourceMap", List["DatasourceMap"], dict, List[dict], None]:
    """Get list of DatasourceLogin objects or dicts. Optionally filter the
    logins by specifying filters.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        default_connection_map: True if connection map is default
            Connection Map. Default False
        project: The project (or its id) which maps are to be fetched.
            Optional unless requesting the default map.
        application: deprecated. Use project instead.
        to_dictionary: If True returns dict, by default (False) returns
            User objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg']

    Returns:
        List["DatasourceMap"] by default
        List[dict] if `to_dict` is True
        DatasourceMap if default map was requested
        dict if default map was requested and `to_dict` is `True`
        None if default map was requested but doesn't exist.

    Examples:
        >>> list_datasource_mappings(connection, name='db_login_name')
    """
    if application:
        helper.deprecation_warning(
            '`application`',
            '`project`',
            '11.3.4.101',  # NOSONAR
            False)
        project = project or application
    return DatasourceMap._list(connection=connection,
                               default_connection_map=default_connection_map, project=project,
                               to_dictionary=to_dictionary, limit=limit, **filters)


class Locale(Entity):
    _OBJECT_TYPE = ObjectTypes.LOCALE
    _API_GETTERS = {
        ('name', 'id'): objects.get_object_info,
    }

    def __init__(self, connection: Connection, id: str):
        """Initialize the Locale object and populate it with I-Server data.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
            id: Locale ID
        """
        if id is None:
            raise AttributeError("Please specify 'id' parameter in the constructor.")
        else:
            super().__init__(connection=connection, object_id=id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)


class DatasourceMap(EntityBase):
    """Object representation of MicroStrategy Datasource Map Object

    The Datasource Map provides mapping between a user and a datasource login
    for the given datasource and datasource connection.

    Attributes:
        connection: A MicroStrategy connection object
        id: The Map's ID.
        project: The project the Map is assigned to.
        default_connection_map: Whether the Map is the default
            for the project.
        ds_connection: The mapped Datasource Connection
        datasource: The mapped Datasource Instance
        user: The mapped User
        login: The mapped Datasource Login
        locale: The Map's locale
    """
    _API_GETTERS: Dict[Union[str, tuple], Callable] = {
        ('id', 'projectId', 'connection', 'datasource', 'user', 'login',
         'locale'): datasources.get_datasource_mapping,
    }
    _API_PATCH: List[Callable] = []
    _FROM_DICT_MAP = {
        "ds_connection": DatasourceConnection.from_dict,
        "datasource": DatasourceInstance.from_dict,
        "user": User.from_dict,
        "login": DatasourceLogin.from_dict,
        "locale": Locale.from_dict
    }

    def __init__(self, connection: Connection, id: Optional[str] = None,
                 project: Optional[Union[Project,
                                         str]] = None, default_connection_map: bool = False,
                 ds_connection: Optional["DatasourceConnection"] = None,
                 datasource: Optional["DatasourceInstance"] = None, user: Optional["User"] = None,
                 login: Optional["DatasourceLogin"] = None, locale: Optional[Locale] = None,
                 application: Optional[Union[Project, str]] = None):
        """Initialise Datasource Map object by passing the ID or by passing
        True for `default_connection_map` and the project for which to
        fetch the default map.

        To explore all available DatasourceMap objects use the
        `list_datasource_mappings()` method.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id: ID of Datasource Map
            project: Project object or ID for restricting the search to
                just this project.
            application: deprecated. Use project instead.
                (Optional except when fetching the default map.)
            default_connection_map: Whether to fetch the project's
                default map.
        """
        if application:
            helper.deprecation_warning(
                '`application`',
                '`project`',
                '11.3.4.101',  # NOSONAR
                False)
            project = project or application
        if id is None:
            if project is None or not default_connection_map:
                helper.exception_handler(
                    "Please either specify the `id` or `default_connection_map` and"
                    " the `project` to find the default map for the given project.")
            else:
                mapping: DatasourceMap = self._list(connection, True, project)
                if mapping:
                    mapping = mapping[0]
                    id = mapping.id
                    ds_connection = mapping.ds_connection
                    datasource = mapping.datasource
                    user = mapping.user
                    login = mapping.login
                    self.__init__(connection, id, project, default_connection_map, ds_connection,
                                  datasource, user, login, locale)
                else:
                    helper.exception_handler(
                        f"The project {project} has no default datasource map.",
                        exception_type=ValueError)
        else:
            super().__init__(connection, id, project=project,
                             default_connection_map=default_connection_map,
                             ds_connection=ds_connection, datasource=datasource, user=user,
                             login=login, locale=locale)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.project: str = kwargs.get("project")
        self.default_connection_map: bool = kwargs.get("default_connection_map")
        self.ds_connection = kwargs.get("ds_connection")
        self.datasource = kwargs.get("datasource")
        self.user = kwargs.get("user")
        self.login = kwargs.get("login")
        self.locale = kwargs.get("locale")

    def delete(self, force: bool = False) -> bool:
        """Deletes the DatasourceMap.

        Note: Default connection maps cannot be deleted.

        Args:
            force: If True, no additional prompt will be shown before deleting.
                Default False.

        Returns:
            True for success. False otherwise.
        """
        user_input = 'N'
        if not force:
            user_input = input(
                (f"Are you sure you want to delete datasource map with ID: {self.id}? [Y/N]: "))
        if force or user_input == 'Y':
            response = datasources.delete_datasource_mapping(self.connection, self.id)
            if response.status_code == 204 and config.verbose:
                print(f"Successfully deleted datasource map with ID: {self.id}")
            return response.ok
        else:
            return False

    @classmethod
    def _list(cls, connection: Connection, default_connection_map: Optional[bool] = False,
              project: Optional[Union[Project, str]] = None, to_dictionary: bool = False,
              limit: Optional[int] = None, **filters) -> Union[List["DatasourceMap"], List[dict]]:
        project_id = project.id if isinstance(project, Project) else project
        try:
            mappings = helper.fetch_objects(connection=connection,
                                            api=datasources.get_datasource_mappings,
                                            default_connection_map=default_connection_map,
                                            project_id=project_id, limit=limit,
                                            dict_unpack_value="mappings", filters=filters)
        except HTTPError as err:
            if err.errno == 404:
                if config.verbose:
                    print("No mapping found.")
                return None
            else:
                raise err
        if to_dictionary:
            return mappings
        else:
            return [cls.from_dict(source=elem, connection=connection) for elem in mappings]

    @classmethod
    def create(cls, connection: Connection, application: Union[Project, str],
               user: Union[User, str], ds_connection: Union[DatasourceConnection, str],
               datasource: Union[DatasourceInstance,
                                 str], login: Union[DatasourceLogin,
                                                    str], locale: Optional[Locale] = None,
               locale_id: Optional[str] = None, locale_name: Optional[str] = None,
               project: Optional[Union[Project, str]] = None) -> "DatasourceMap":
        """Create a new Datasource Map object on the server.
        If more than one locale related parameters are provided,
        `locale` has priority, then `locale_id`.

        Args:
            connection: A MicroStrategy connection object
            application: The project the Map is to be assigned to. Will be
            removed from 11.3.4.101 and replaced with project.
            project: Use from 11.3.4.101 instead of application.
            user: The User to be mapped
            ds_connection: The Datasource Connection to be mapped
            datasource: The Datasource Instance to be mapped
            login: The Datasource Login to be mapped
            locale: The locale to be maped.
            locale_id: The id of locale to be maped.
            locale_name: The name of locale to be maped.

        Returns:
            DatasourceMap object
        """
        helper.deprecation_warning(
            '`application`',
            '`project`',
            '11.3.4.101',  # NOSONAR
            False,
            False)
        project = project or application
        project_id = get_objects_id(project, Project)
        user_id = get_objects_id(user, User)
        connection_id = get_objects_id(ds_connection, DatasourceConnection)
        datasource_id = get_objects_id(datasource, DatasourceInstance)
        login_id = get_objects_id(login, DatasourceLogin)
        body = {
            "projectId": project_id,
            "user": {
                "id": user_id
            },
            "connection": {
                "id": connection_id
            },
            "datasource": {
                "id": datasource_id
            },
            "login": {
                "id": login_id
            },
        }
        if locale and isinstance(locale, Locale):
            body["locale"] = locale.to_dict()
        elif locale_id and isinstance(locale_id, str):
            body["locale"] = {"id": locale}
        elif locale_name and isinstance(locale_name, str):
            body["locale"] = {"name": locale}

        jresponse = datasources.create_datasource_mapping(connection=connection, body=body).json()
        if config.verbose:
            print("Successfully created datasource connection map "
                  f"with ID: '{jresponse.get('id')}'")
        return cls.from_dict(source=helper.camel_to_snake(jresponse), connection=connection)

    # TODO: improve to/from dict methods to allow renaming
    # on _FROM_DICT_MAP level.

    @classmethod
    def from_dict(cls, source: Dict[str, Any], connection: Connection) -> "DatasourceMap":

        def translate_names(name: str):
            if name == "connection":
                return "ds_connection"
            elif name == "project_id":
                return "project"
            else:
                return name

        modified_source = {translate_names(key): val for key, val in source.items()}
        return super(EntityBase, cls).from_dict(modified_source, connection)

    def to_dict(self, camel_case):

        def translate_names(name: str):
            if name == "ds_connection":
                return "connection"
            elif name == "project":
                return "project_id"
            else:
                return name

        thedict = super().to_dict(camel_case=camel_case)
        return {translate_names(key): val for key, val in thedict.items()}



from functools import wraps
import inspect
from typing import Any, Callable, Union

from requests.adapters import Response

from mstrio import config
from mstrio.api.exceptions import MstrException, PartialSuccess, Success
from mstrio.utils.helper import get_default_args_from_func, response_handler


def get_args_and_bind_values(func: Callable[[Any], Any], *args, **kwargs):
    signature = inspect.signature(func)
    return signature.bind(*args, **kwargs).arguments


class ErrorHandler:
    """An easy-to-use class decorator designed to replace
    logic responsible for displaying the error message in API wrappers.

    Attributes:
        err_msg(str): error message to be displayed in case of error

    Usage:
        to replace the code below

          if not response.ok:
            if error_msg is None:
              error_msg = f'Error deleting Datasource Login with ID {id}'
            response_handler(response, error_msg)
          return response

        use the decorator in a following way

           @ErrorHandler(err_msg='Error deleting Datasource Login with ID {id}')
           def func(connection, id):
              ...

        the strings in curly braces will be replaced
        with the appropriate values if they appear in the function arguments
    """

    def __init__(self, err_msg: str):
        self._err_msg = err_msg

    def __call__(self, func: Callable[[Any], Any]):

        @wraps(func)
        def inner(*args, **kwargs):
            response = func(*args, **kwargs)
            error_msg = kwargs.get("error_msg") if kwargs.get("error_msg") else self._err_msg
            if not response.ok:
                handler_kwargs = self._get_resp_handler_kwargs(kwargs)
                error_msg = self._replace_with_values(error_msg, func, *args, **kwargs)
                response_handler(response, error_msg, **handler_kwargs)
            return response

        return inner

    @staticmethod
    def _replace_with_values(err_msg: str, func: Callable[[Any], Any], *args, **kwargs):
        all_args = get_args_and_bind_values(func, *args, **kwargs)
        for arg in all_args:
            arg_name, arg_value = arg, all_args[arg]
            err_msg = err_msg.replace(f'{{{arg_name}}}', str(arg_value))
        return err_msg

    @staticmethod
    def _get_resp_handler_kwargs(decorated_func_kwargs):
        default_args = get_default_args_from_func(response_handler)
        for arg in default_args:
            default_args[arg] = decorated_func_kwargs.get(arg, default_args[arg])
        return default_args


def bulk_operation_response_handler(
        response: Response,
        unpack_value: str = None) -> Union[PartialSuccess, Success, MstrException]:
    """Handle partial success and other statuses from bulk operation."""
    response_body = response.json()
    if response.ok and unpack_value:
        response_body = response_body[unpack_value]

    if response.status_code == 200:
        err = Success(response_body)
    elif response.status_code == 207:
        err = PartialSuccess(response_body)
    else:
        err = MstrException(response_body)

    if config.verbose:
        print(err)
    return err



import json

from mstrio import config
from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.helper import response_handler


def list_schedules(connection, fields=None, error_msg=None):
    """Get list of a schedules.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """

    response = connection.session.get(
        url=f'{connection.base_url}/api/schedules',
        params={'fields': fields}
    )
    if config.debug:
        print(response.url)
    if not response.ok:
        if error_msg is None:
            error_msg = "Error getting schedules list."
        response_handler(response, error_msg)

    # Fix for incorrect 'eventId' (expecting 'id')
    event_based_in_list = False
    response_json = response.json()
    for schedule in response_json['schedules']:
        if 'event' in schedule.keys():
            schedule['event']['id'] = schedule['event'].pop('eventId')
            event_based_in_list = True
    if event_based_in_list:
        response.encoding, response._content = 'utf-8', json.dumps(response_json).encode('utf-8')

    return response


def get_schedule(connection, id, fields=None, error_msg=None):
    """Get information of a specific schedule by `schedule_id`.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        schedule_id(str): ID of the schedule
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """

    response = connection.session.get(
        url=f'{connection.base_url}/api/schedules/{id}',
        params={'fields': fields}
    )
    if config.debug:
        print(response.url)
    if not response.ok:
        if error_msg is None:
            error_msg = "Error getting schedule information."
        response_handler(response, error_msg)

    # Fix for incorrect 'eventId' (expecting 'id')
    response_json = response.json()
    if 'event' in response_json.keys():
        response_json['event']['id'] = response_json['event'].pop('eventId')
        response.encoding, response._content = 'utf-8', json.dumps(response_json).encode('utf-8')

    return response


def create_schedule(connection, body, fields=None, error_msg=None):
    """Create a new schedule using data from `body`.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        body(dict): Dictionary containging data used for creating schedule.
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """

    # id to eventId conversion - API Problem
    if 'event' in body.keys():
        body['event']['eventId'] = body['event'].pop('id')

    response = connection.session.post(
        url=f'{connection.base_url}/api/schedules',
        json=body,
        params={'fields': fields}
    )

    if config.debug:
        print(response.url)
    if not response.ok:
        if error_msg is None:
            error_msg = "Error getting schedule information."
        response_handler(response, error_msg)

    # Fix for incorrect 'eventId' (expecting 'id')
    response_json = response.json()
    if 'event' in response_json.keys():
        response_json['event']['id'] = response_json['event'].pop('eventId')
        response.encoding, response._content = 'utf-8', json.dumps(response_json).encode('utf-8')

    return response


def update_schedule(connection, id, body, fields=None, error_msg=None):
    """Alter a schedule specified by `schedule_id`, using data from `body`.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        schedule_id(str): ID of the schedule
        body(dict): Dictionary containging data used for replacing schedule.
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """

    # id to eventId conversion - API Problem
    if 'event' in body.keys():
        body['event']['eventId'] = body['event'].pop('id')

    response = connection.session.put(
        url=f'{connection.base_url}/api/schedules/{id}',
        json=body,
        params={'fields': fields}
    )
    if config.debug:
        print(response.url)
    if not response.ok:
        if error_msg is None:
            error_msg = "Error getting schedule information."
        response_handler(response, error_msg)

    # Fix for incorrect 'eventId' (expecting 'id')
    response_json = response.json()
    if 'event' in response_json.keys():
        response_json['event']['id'] = response_json['event'].pop('eventId')
    response.encoding, response._content = 'utf-8', json.dumps(response_json).encode('utf-8')

    return response


@ErrorHandler(err_msg='Error getting schedule {id} information.')
def delete_schedule(connection, id, fields=None, error_msg=None):
    """Delete a schedule specified by `schedule_id`.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id(str): ID of the schedule
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    response = connection.session.delete(
        url=f'{connection.base_url}/api/schedules/{id}',
        params={'fields': fields}
    )

    if config.debug:
        print(response.url)

    return response



from typing import List, TYPE_CHECKING, Union

from mstrio import config
from mstrio.api import datasources, objects
from mstrio.users_and_groups.user import User
from mstrio.utils import helper
from mstrio.utils.entity import DeleteMixin, CopyMixin, Entity, ObjectTypes

if TYPE_CHECKING:
    from mstrio.connection import Connection


def list_datasource_logins(connection: "Connection", to_dictionary: bool = False,
                           limit: int = None,
                           **filters) -> Union[List["DatasourceLogin"], List[dict]]:
    """Get list of DatasourceLogin objects or dicts. Optionally filter the
    logins by specifying filters.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        to_dictionary: If True returns dict, by default (False) returns
            User objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg']

    Examples:
        >>> list_db_logins(connection, name='db_login_name')
    """
    return DatasourceLogin._list_datasource_logins(
        connection=connection,
        to_dictionary=to_dictionary,
        limit=limit,
        **filters,
    )


class DatasourceLogin(Entity, CopyMixin, DeleteMixin):
    """A user login configuration object to access a particular datasource. Also
    formerly known as database login.

    Attributes:
        name: name of the database login
        id: database login ID
        description: Description of the database login
        username: database user to be used by the database login
        version: Object version ID
        ancestors: List of ancestor folders
        type: Object type. Enum
        subtype: Object subtype
        ext_type: Object extended type.
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        owner: User object that is the owner
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """
    _OBJECT_TYPE = ObjectTypes.DBLOGIN
    _FROM_DICT_MAP = {**Entity._FROM_DICT_MAP, 'owner': User.from_dict}
    _API_GETTERS = {
        ('abbreviation', 'type', 'subtype', 'ext_type', 'version', 'owner', 'icon_path',
         'view_media', 'ancestors', 'certified_info', 'acl'): objects.get_object_info,
        ('id', 'name', 'description', 'username', 'date_created', 'date_modified',
         'acg'): datasources.get_datasource_login
    }
    _API_PATCH: dict = {
        ("abbreviation"): (objects.update_object, "partial_put"),
        ("name", "username", "description", "password"):
            (datasources.update_datasource_login, "patch")
    }
    _PATCH_PATH_TYPES = {"name": str, "username": str, "description": str, "password": str}

    def __init__(self, connection: "Connection", name: str = None, id: str = None):
        """Initialize DatasourceLogin object."""

        if id is None and name is None:
            raise ValueError("Please specify either 'id' or 'name' parameter in the constructor.")

        if id is None:
            objects_info = DatasourceLogin._list_datasource_logins(connection=connection,
                                                                   name=name, to_dictionary=True)
            if objects_info:
                id, object_info = objects_info[0].pop("id"), objects_info[0]
                super().__init__(connection=connection, object_id=id, **object_info)
            else:
                raise ValueError(f"There is no Datasource Login: '{name}'")
        else:
            super().__init__(connection=connection, object_id=id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.username = kwargs.get("username")

    def alter(self, name: str = None, username: str = None, description: str = None,
              password: str = None) -> None:
        """Alter the datasource login properties.

        Args:
            name: login object name
            username: username
            description: login object description
            password: database password to be used by the database login
        """
        func = self.alter
        args = func.__code__.co_varnames[:func.__code__.co_argcount]
        defaults = func.__defaults__  # type: ignore
        default_dict = dict(zip(args[-len(defaults):], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    @classmethod
    def create(cls, connection: "Connection", name: str, username: str, password: str,
               description: str = None) -> "DatasourceLogin":
        """Create a new datasource login.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            name: login object name
            username: username
            description: login object description
            password: database password to be used by the database login
        Returns:
            DatasourceConnection object.
        """
        body = {
            "name": name,
            "description": description,
            "username": username,
            "password": password
        }
        body = helper.delete_none_values(body)
        response = datasources.create_datasource_login(connection, body).json()
        if config.verbose:
            print("Successfully created datasource login named: '{}' with ID: '{}'".format(
                response.get('name'), response.get('id')))
        return cls.from_dict(source=response, connection=connection)

    @classmethod
    def _list_datasource_logins(cls, connection: "Connection", to_dictionary: bool = False,
                                limit: int = None,
                                **filters) -> Union[List["DatasourceLogin"], List[dict]]:
        objects = helper.fetch_objects(connection=connection,
                                       api=datasources.get_datasource_logins,
                                       dict_unpack_value="logins", limit=limit, filters=filters)
        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]



# flake8: noqa
from . import datasets
from .document import Document, list_documents, list_documents_across_projects
from .dossier import Dossier, list_dossiers, list_dossiers_across_projects
from .library import Library
from .report import list_reports, Report



from mstrio.utils.error_handlers import ErrorHandler
from mstrio.connection import Connection
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from requests_futures.sessions import FuturesSession


@ErrorHandler(err_msg='Error while creating the folder.')
def create_folder(connection: Connection, name: str, parent_id: str,
                  description: Optional[str] = None, project_id: Optional[str] = None):
    """Create a folder.

    Note:
        When `project_id` is provided then folder will be created in this
        project. Otherwise it will be created in a project selected within
        `connection` object.

    Args:
        connection: MicroStrategy REST API connection object
        name (string): name of folder to create
        parent_id (string): id of folder in which new folder will be created
        description (string, optional): description of folder to create
        project_id (string, optional): id of project

    Returns:
        Complete HTTP response object.
    """
    project_id = project_id if project_id is not None else connection.project_id
    body = {
        "name": name,
        "description": description,
        "parent": parent_id,
    }
    return connection.session.post(url=connection.base_url + '/api/folders',
                                   headers={'X-MSTR-ProjectID': project_id}, json=body)


@ErrorHandler(err_msg='Error while deleting folder with ID: {id}.')
def delete_folder(connection: Connection, id: str, project_id: Optional[str] = None):
    """Delete complete folder.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of folder
        project_id (string, optional): id of project

    Returns:
        Complete Future object.
    """
    project_id = project_id if project_id is not None else connection.project_id
    return connection.session.delete(url=f"{connection.base_url}/api/folders/{id}",
                                     headers={'X-MSTR-ProjectID': project_id})


@ErrorHandler(err_msg='Error while listing folders.')
def list_folders(connection: Connection, project_id: Optional[str] = None, offset: int = 0,
                 limit: int = 5000, error_msg: Optional[str] = None):
    """Get a list of folders.

    Args:
        connection: MicroStrategy REST API connection object
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete HTTP response object.
    """
    return connection.session.get(url=f"{connection.base_url}/api/folders",
                                  headers={'X-MSTR-ProjectID': project_id}, params={
                                      'offset': offset,
                                      'limit': limit
                                  })


def list_folders_async(future_session: "FuturesSession", connection: Connection,
                       project_id: Optional[str] = None, offset: int = 0, limit: int = 5000):
    """Get a list of folders asynchronously.

    Args:
        future_session(object): `FuturesSession` object to call MicroStrategy
            REST Server asynchronously
        connection: MicroStrategy REST API connection object
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete Future object.
    """
    url = f"{connection.base_url}/api/folders"
    headers = {'X-MSTR-ProjectID': project_id}
    params = {'offset': offset, 'limit': limit}
    return future_session.get(url=url, headers=headers, params=params)


@ErrorHandler(err_msg='Error while getting contents of a folder with ID: {id}.')
def get_folder_contents(connection: Connection, id: str, project_id: Optional[str] = None,
                        offset: int = 0, limit: int = 5000, error_msg: Optional[str] = None):
    """Get contents of a folder.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of folder
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete HTTP response object.
    """
    project_id = project_id if project_id is not None else connection.project_id
    return connection.session.get(url=f"{connection.base_url}/api/folders/{id}",
                                  headers={'X-MSTR-ProjectID': connection.project_id}, params={
                                      'offset': offset,
                                      'limit': limit
                                  })


def get_folder_contents_async(future_session: "FuturesSession", connection: Connection, id: str,
                              project_id: Optional[str] = None, offset: int = 0,
                              limit: int = 5000):
    """Get contents of a folder asynchronously.

    Args:
        future_session(object): `FuturesSession` object to call MicroStrategy
            REST Server asynchronously
        connection: MicroStrategy REST API connection object
        id (string): ID of folder
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete Future object.
    """
    project_id = project_id if project_id is not None else connection.project_id
    url = f"{connection.base_url}/api/folders/{id}"
    headers = {'X-MSTR-ProjectID': project_id}
    params = {'offset': offset, 'limit': limit}
    return future_session.get(url=url, headers=headers, params=params)


@ErrorHandler(err_msg='Error while getting contents of a pre-defined folder.')
def get_predefined_folder_contents(connection: Connection, folder_type: int,
                                   project_id: Optional[str] = None, offset: int = 0,
                                   limit: int = 5000, error_msg: Optional[str] = None):
    """Get contents of a pre-defined folder.

    Args:
        connection: MicroStrategy REST API connection object
        folder_type (int): predefined folder type, from `EnumDSSXMLFolderNames`
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete HTTP response object.
    """
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    return connection.session.get(
        url=f"{connection.base_url}/api/folders/preDefined/{folder_type}",
        headers={'X-MSTR-ProjectID': project_id}, params={
            'offset': offset,
            'limit': limit
        })


def get_predefined_folder_contents_async(future_session: "FuturesSession", connection: Connection,
                                         folder_type: int, project_id: Optional[str] = None,
                                         offset: int = 0, limit: int = 5000):
    """Get contents of a pre-defined folder.

    Args:
        future_session(object): `FuturesSession` object to call MicroStrategy
            REST Server asynchronously
        connection: MicroStrategy REST API connection object
        folder_type (int): predefined folder type, from `EnumDSSXMLFolderNames`
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete Future object.
    """
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    url = f"{connection.base_url}/api/folders/preDefined/{folder_type}"
    headers = {'X-MSTR-ProjectID': project_id}
    params = {'offset': offset, 'limit': limit}
    return future_session.get(url=url, headers=headers, params=params)


@ErrorHandler(err_msg='Error while getting contents of My Personal Objects folder.')
def get_my_personal_objects_contents(connection: Connection, project_id: Optional[str] = None):
    """Get contents of My Personal Objects folder.

    Args:
        connection: MicroStrategy REST API connection object
        project_id (string, optional): id of project

    Returns:
         Complete HTTP response object.
    """
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    return connection.session.get(url=f"{connection.base_url}/api/folders/myPersonalObjects",
                                  headers={'X-MSTR-ProjectID': project_id})



from typing import List, Union

from mstrio import config
from mstrio.api import events, objects
from mstrio.connection import Connection
from mstrio.utils import helper
from mstrio.utils.entity import Entity, ObjectTypes
from packaging import version

from ..utils.wip import module_wip, WipLevels

module_wip(globals(), level=WipLevels.WARNING)


def list_events(connection: Connection, to_dictionary: bool = False, limit: int = None,
                **filters) -> Union[List["Event"], List[dict]]:
    """List event objects or event dictionaries. Optionally filter list.

    Args:
        connection(object): MicroStrategy connection object returned
            by 'connection.Connection()'
        to_dictionary(bool, optional): if True, return event as
            list of dicts
        limit(int, optional): maximum number of Events returned.
        **filters: Available filter parameters:['name':,
                                                'id',
                                                'description']
    """
    _objects = helper.fetch_objects(connection=connection, api=events.list_events, limit=limit,
                                    filters=filters, dict_unpack_value='events')

    if to_dictionary:
        return _objects
    else:
        return [Event.from_dict(source=obj, connection=connection) for obj in _objects]


class Event(Entity):
    """Class representation of MicroStrategy Event object.

    Attributes:
        connection: A MicroStrategy connection object
        name: Event name
        id: Event ID
        description: Event descriptions
    """
    _PATCH_PATH_TYPES = {'name': str, 'description': str}
    _OBJECT_TYPE = ObjectTypes.SCHEDULE_EVENT
    _API_GETTERS = {
        ('abbreviation', 'type', 'subtype', 'ext_type', 'date_created', 'date_modified', 'version',
         'owner', 'icon_path', 'view_media', 'ancestors', 'certified_info', 'acg',
         'acl'): objects.get_object_info,
    }

    def __init__(self, connection: Connection, id: str = None, name: str = None):
        """Initialize the Event object, populates it with I-Server data.
        Specify either `id` or `name`. When `id` is provided (not `None`),
        `name` is omitted.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
            id: Event ID
            name: Event name
        """
        self._API_GETTERS[('id', 'name', 'description')] = \
            events.get_event if version.parse(connection.web_version) >= \
            version.parse('11.3.0200') else objects.get_object_info

        if id is None and name is None:
            raise AttributeError(
                "Please specify either 'name' or 'id' parameter in the constructor.")
        if id is None:
            objects_info = list_events(connection, name=name, to_dictionary=True)
            if objects_info:
                id, object_info = objects_info[0].pop("id"), objects_info[0]
                super().__init__(connection=connection, object_id=id, **object_info)
            else:
                raise ValueError(f"There is no event with the given name: '{name}'")
        else:
            super().__init__(connection=connection, object_id=id, name=name)

    def trigger(self):
        """Triggers the Event
        NOTE: WIP, may not work properly"""
        response = events.trigger_event(self.connection, self.id)
        if response.ok and config.verbose:
            print(f'Event \'{self.name}\' with ID : \'{self.id}\' has been triggered.')
        return response.ok



from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg='Error triggering event {id}')
def trigger_event(connection, id, fields=None, error_msg=None):
    """Triggers an event.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id(str): ID of the event
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    url = f'{connection.base_url}/api/events/{id}/trigger'
    return connection.session.post(url=url, params={'fields': fields})


@ErrorHandler(err_msg='Error getting events list.')
def list_events(connection, fields=None, error_msg=None):
    """Get list of a events.

    Args:
        connection(object): MicroStrategy connection object returned by
                `connection.Connection()`.
        error_msg(str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.session.get(url=f'{connection.base_url}/api/events',
                                  params={'fields': fields})


@ErrorHandler(err_msg='Error getting event {id} information.')
def get_event(connection, id, fields=None, error_msg=None):
    """Get information of a specific event

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id(str): ID of the event
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.session.get(
        url=f'{connection.base_url}/api/events/{id}',
        params={'fields': fields},
    )



from enum import Enum, IntFlag
from typing import Any, Dict, List, Optional, TYPE_CHECKING, TypeVar, Union
from mstrio.api import objects
from mstrio.utils import helper
from mstrio.utils.helper import Dictable
import pandas as pd
from requests import HTTPError
from mstrio.connection import Connection
from mstrio.types import ObjectTypes

if TYPE_CHECKING:
    from mstrio.users_and_groups import UserOrGroup
    from mstrio.utils.entity import Entity
    from mstrio.server import Project


class Rights(IntFlag):
    """"Enumeration constants used to specify the access granted attribute of
    the DSS objects. """
    EXECUTE = 0b10000000
    USE = 0b01000000
    CONTROL = 0b00100000
    DELETE = 0b00010000
    WRITE = 0b00001000
    READ = 0b00000100
    USE_EXECUTE = 0b00000010  # This constant is deprecated
    BROWSE = 0b00000001
    INHERITABLE = 0b100000000000000000000000000000


class Permissions(str, Enum):
    """Enumeration constants used to specify combination of Rights values
    similar to workstation Security Access.

    TODO: This has to be string-based to discern between 'Denied All'
    and 'Full Control', which have the same mask.
    """
    DENIED_ALL = 'Denied All'
    DEFAULT_ALL = 'Default All'
    CONSUME = 'Consume'
    VIEW = 'View'
    MODIFY = 'Modify'
    FULL_CONTROL = 'Full Control'


class AggregatedRights(IntFlag):
    """Enumeration constants used to specify combination of Rights values."""
    NONE = 0b00000000
    CONSUME = 0b01000101
    VIEW = 0b11000101
    MODIFY = 0b11011101
    ALL = 0b11111111


AGGREGATED_RIGHTS_MAP = {
    Permissions.VIEW: AggregatedRights.VIEW,
    Permissions.MODIFY: AggregatedRights.MODIFY,
    Permissions.FULL_CONTROL: AggregatedRights.ALL,
    Permissions.DENIED_ALL: AggregatedRights.ALL,
    Permissions.DEFAULT_ALL: AggregatedRights.NONE,
    Permissions.CONSUME: AggregatedRights.CONSUME,
}

T = TypeVar("T")


class ACE(Dictable):

    _FROM_DICT_MAP = {
        'rights': Rights,
    }

    def __init__(self, deny: bool, entry_type: int, rights: Rights, trustee_id: str,
                 trustee_name: str, trustee_type: int, trustee_subtype: int, inheritable: bool):
        """Set ACL object.

        Args:
            deny: Specifies whether access is denied
            entry_type: Access control entry type (1 for object access).
                Possible values can be found in EnumDSSXMLAccessEntryType
            rights: Rights assigned to the designated trustee
            trustee_id: User ID of the designated trustee
            trustee_name: User name of the designated trustee
            trustee_type: Type of the designated trustee
            trustee_subtype: Sub-type of the designated trustee
            inheritable: Specifies whether access control is inherited
        """

        self.deny = deny
        self.entry_type = entry_type
        self.rights = rights
        self.trustee_id = trustee_id
        self.trustee_name = trustee_name
        self.trustee_type = trustee_type
        self.trustee_subtype = trustee_subtype
        self.inheritable = inheritable

    def __eq__(self, other: object) -> bool:
        if len(self.__dict__) != len(other.__dict__):
            return False
        for attr in self.__dict__:
            if getattr(self, attr) != getattr(other, attr):
                return False
        return True

    @classmethod
    def from_dict(cls, source: Dict[str, Any], connection: Connection):

        def translate_names(name: str):
            if name == "type":
                return "entry_type"
            return name

        modified_source = {translate_names(key): val for key, val in source.items()}
        return super().from_dict(modified_source, connection)

    def to_dict(self, camel_case=True):

        def translate_names(name: str):
            if name == "entry_type" or name == "entryType":
                return "type"
            return name

        result_dict = super().to_dict(camel_case=camel_case)
        return {translate_names(key): val for key, val in result_dict.items()}


class ACLMixin:
    """ACLMixin class adds Access Control List (ACL) management for supporting
    objects.

    An ACL is a set of permissions on objects so that users or user groups have
    control over individual objects in the system. Those permissions decide
    whether or not a user can perform a particular class of operations on a
    particular object. For example, a user may have permissions to view and
    execute a report , but cannot modify the report definition or delete the
    report.

    NOTE: Must be mixedin with Entity or its subclasses.
    """

    def list_acl(self, to_dataframe: bool = False, **filters) -> Union[pd.DataFrame, list]:
        """Get Access Control List (ACL) for this object. Optionally filter
        ACLs by specifying filters.

        Args:
            to_dataframe(bool, optional): if True, return datasets as
                pandas DataFrame
            **filters: Available filter parameters: [deny, type, rights,
                trustee_id, trustee_name, trustee_type, trustee_subtype,
                inheritable]

        Examples:
            >>> list_acl(deny=True, trustee_name="John")
        """
        acl = helper.filter_obj_list(self.acl, **filters)
        if to_dataframe:
            return pd.DataFrame(acl)
        else:
            return acl

    def acl_add(self: "Entity", rights: Union[int, Rights, AggregatedRights],
                trustees: Union["UserOrGroup", List["UserOrGroup"]], denied: bool = False,
                inheritable: bool = False, propagate_to_children: Optional[bool] = None) -> None:
        """Add Access Control Element (ACE) to the object ACL.

        Note:
            Argument `propagate_to_children` is used only for objects with
            type `ObjectTypes.FOLDER`.

        Args:
            rights: The degree to which the user or group is granted or denied
                access to the object. The available permissions are defined in
                `Rights` and `AggregatedRights` Enums
            trustees: list of trustees (`User` or `UserGroup` objects or ids) to
                update the ACE for
            denied: flag to indicate granted or denied access to the object
            inheritable: Applies only to folders. If set, any objects placed in
                the folder inherit the folder's entry in the ACL.
            propagate_to_children: used for folder objects only, default value
                is None, if set to True/False adds `propagateACLToChildren`
                keyword to the request body and sets its value accordingly

        Examples:
            >>> obj.acl_add(rights=Rights.BROWSE | Rights.EXECUTE,
            >>>             trustees=user_obj, denied=True)
        """
        self._update_acl(op="ADD", rights=rights, trustees=trustees, denied=denied,
                         inheritable=inheritable, propagate_to_children=propagate_to_children)

    def acl_remove(self: "Entity", rights: Union[int, Rights, AggregatedRights],
                   trustees: Union["UserOrGroup", List["UserOrGroup"]], denied: bool = False,
                   inheritable: bool = False,
                   propagate_to_children: Optional[bool] = None) -> None:
        """Remove Access Control Element (ACE) from the object ACL.

        Note:
            Argument `propagate_to_children` is used only for objects with
            type `ObjectTypes.FOLDER`.

        Args:
            rights: The degree to which the user or group is granted or denied
                access to the object. The available permissions are defined in
                `Rights` and `AggregatedRights` Enums
            trustees: list of trustees (`User` or `UserGroup` objects or ids) to
                update the ACE for
            denied: flag to indicate granted or denied access to the object
            inheritable: Applies only to folders. If set, any objects placed in
                the folder inherit the folder's entry in the ACL.
            propagate_to_children: used for folder objects only, default value
                is None, if set to True/False adds `propagateACLToChildren`
                keyword to the request body and sets its value accordingly

        Examples:
            >>> obj.acl_remove(rights=Rights.BROWSE | Rights.EXECUTE,
            >>>                trustees=user_obj, denied=True)
        """
        self._update_acl(op="REMOVE", rights=rights, trustees=trustees, denied=denied,
                         inheritable=inheritable, propagate_to_children=propagate_to_children)

    def acl_alter(self: "Entity", rights: Union[int, Rights, AggregatedRights],
                  trustees: Union["UserOrGroup", List["UserOrGroup"]], denied: bool = False,
                  inheritable: bool = False, propagate_to_children: Optional[bool] = None) -> None:
        """Alter an existing Access Control Element (ACE) of the object ACL.

        Note:
            Argument `propagate_to_children` is used only for objects with
            type `ObjectTypes.FOLDER`.

        Args:
            rights: The degree to which the user or group is granted or denied
                access to the object. The available permissions are defined in
                `Rights` and `AggregatedRights` Enums
            trustees: list of trustees (`User` or `UserGroup` objects or ids) to
                update the ACE for
            denied: flag to indicate granted or denied access to the object
            inheritable: Applies only to folders. If set, any objects placed in
                the folder inherit the folder's entry in the ACL.
            propagate_to_children: used for folder objects only, default value
                is None, if set to True/False adds `propagateACLToChildren`
                keyword to the request body and sets its value accordingly

        Examples:
            >>> obj.acl_alter(rights=Rights.BROWSE | Rights.EXECUTE,
            >>>               trustees=user_obj, denied=True)
        """
        self._update_acl(op="REPLACE", rights=rights, trustees=trustees, denied=denied,
                         inheritable=inheritable, propagate_to_children=propagate_to_children)

    def _update_acl(self: "Entity", op: str, rights: Union[int, Rights, AggregatedRights],
                    trustees: Union["UserOrGroup", List["UserOrGroup"]],
                    propagate_to_children: Optional[bool] = None, denied: bool = False,
                    inheritable: bool = False) -> None:
        """Updates the ACL for this object, performs operation defined by the
        `op` parameter on all objects from `trustees` list.

        Note:
            Argument `propagate_to_children` is used only for objects with
            type `ObjectTypes.FOLDER`.

        Args:
            op: ACL update operator, available values are "ADD", "REMOVE" and
                "REPLACE"
            rights: value of rights to use by the operator
            trustees: list of trustees to update the ACE for
            propagate_to_children: used for folder objects only, default value
                is None, if set to True/False adds `propagateACLToChildren`
                keyword to the request body and sets its value accordingly
            denied: flag to indicate granted or denied access to the object
            inheritable: Applies only to folders. If set, any objects placed in
                the folder inherit the folder's entry in the ACL.
        """
        # TODO merge duplicated code with _modify_rights function
        # TODO move (op, rights, ids, propagate_to_children=None,
        # denied=None, inheritable=None, types=None) to
        # separate AccesControlEntry class
        if op not in ["ADD", "REMOVE", "REPLACE"]:
            raise ValueError("Wrong ACL operator passed. Please use ADD, REMOVE or REPLACE")

        # convert Enums
        rights = rights.value if isinstance(rights, Enum) else rights
        if rights not in range(256) and rights not in range(536_870_912, 536_871_167):
            msg = ("Wrong `rights` value, please provide value in range 0-255 or combination of "
                   "Rights enums")
            raise ValueError(msg)
        if not isinstance(trustees, list):
            trustees = [trustees]
        trustee_ids = [trustee if isinstance(trustee, str) else trustee.id for trustee in trustees]
        body = {
            "acl": [
                {
                    "op": op,
                    "trustee": id,
                    "rights": rights,
                    "type": 1,  # EnumDSSXMLAccessEntryType
                    "denied": denied,
                    "inheritable": inheritable
                } for id in trustee_ids
            ]
        }

        if isinstance(propagate_to_children, bool) and self._OBJECT_TYPE is ObjectTypes.FOLDER:
            body["propagateACLToChildren"] = propagate_to_children

        response = objects.update_object(connection=self.connection, id=self.id, body=body,
                                         object_type=self._OBJECT_TYPE.value)
        if response.ok:
            response = response.json()
            self._set_object(**response)


class TrusteeACLMixin:
    """TrusteeACLMixin class adds ACL management for Trustee classes.

    Objects currently supporting this Mixin are: (`User` and `UserGroup`).
    """

    def set_permission(self, permission: Permissions, to_objects: Union[str, List[str]],
                       object_type: "ObjectTypes", project: Optional[Union[str, "Project"]] = None,
                       propagate_to_children: Optional[bool] = None,
                       application: Optional[Union[str, "Project"]] = None) -> None:
        """Set permission to perform actions on given object(s).

        Function is used to set permission of the trustee to perform given
        actions on the provided objects. Within one execution of the function
        permission will be set in the same manner for each of the provided
        objects. The only available values of permission are: 'View', 'Modify',
        'Full Control', 'Denied All', 'Default All'. Permission is the
        predefined set of rights. All objects to which the rights will be given
        have to be of the same type which is also provided.

        Args:
            permission: The Permission which defines set of rights. See:
                `Permissions` enum
            to_objects: List of object ids on access list for which the
                permissions will be set
            object_type: Type of objects on access list. See: `ObjectTypes` enum
            project: Object or id of Project where the object is
                located. If not passed, Project (project_id) selected in
                Connection object is used
            application: deprecated. Use project instead.
            propagate_to_children: Flag used in the request to determine if
                those rights will be propagated to children of the trustee
        Returns:
            None
        """
        if application:
            helper.deprecation_warning(
                '`application`',
                '`project`',
                '11.3.4.101',  # NOSONAR
                False)
            project = project or application

        if not isinstance(permission, Permissions):
            try:
                permission = Permissions(permission)
            except ValueError:
                msg = ("Invalid `permission` value. Available values are: 'View', "
                       "'Modify', 'Full Control', 'Denied All', 'Default All'. "
                       "See: Permissions enum.")
                helper.exception_handler(msg)
        right_value = AGGREGATED_RIGHTS_MAP[permission].value
        denied = permission is Permissions.DENIED_ALL

        # those 2 tries are for clearing current rights (set to default values)
        try:
            _modify_rights(connection=self.connection, trustee_id=self.id, op='REMOVE',
                           rights=AggregatedRights.ALL.value, ids=to_objects,
                           object_type=object_type, project=project, denied=(not denied),
                           propagate_to_children=propagate_to_children, verbose=False)
        except HTTPError:
            pass
        try:
            _modify_rights(connection=self.connection, trustee_id=self.id, op='REMOVE',
                           rights=AggregatedRights.ALL.value, ids=to_objects,
                           object_type=object_type, project=project, denied=denied,
                           propagate_to_children=propagate_to_children, verbose=False)
        except HTTPError:
            pass

        if not permission == Permissions.DEFAULT_ALL:
            _modify_rights(connection=self.connection, trustee_id=self.id, op='ADD',
                           rights=right_value, ids=to_objects, object_type=object_type,
                           project=project, denied=denied,
                           propagate_to_children=propagate_to_children)

    def set_custom_permissions(self, to_objects: Union[str, List[str]], object_type: "ObjectTypes",
                               project: Union[str,
                                              "Project"] = None, execute: Optional[str] = None,
                               use: Optional[str] = None, control: Optional[str] = None,
                               delete: Optional[str] = None, write: Optional[str] = None,
                               read: Optional[str] = None, browse: Optional[str] = None,
                               application: Union[str, "Project"] = None) -> None:
        if application:
            helper.deprecation_warning(
                '`application`',
                '`project`',
                '11.3.4.101',  # NOSONAR
                False)
            project = project or application
        """Set custom permissions to perform actions on given object(s).

        Function is used to set rights of the trustee to perform given actions
        on the provided objects. Within one execution of the function rights
        will be set in the same manner for each of the provided objects.
        None of the rights is necessary, but if provided then only possible
        values are 'grant' (to grant right), 'deny' (to deny right), 'default'
        (to reset right) or None which is default value and means that nothing
        will be changed for this right. All objects to which the rights will be
        given have to be of the same type which is also provided.

        Args:
            to_objects: (str, list(str)): List of object ids on access list to
                which the permissions will be set
            object_type (int): Type of objects on access list
            project (str, Project): Object or id of Project in which
                the object is located. If not passed, Project
                (project_id) selected in Connection object is used.
            application: deprecated. Use project instead.
            execute (str): value for right "Execute". Available are 'grant',
                'deny', 'default' or None
            use (str): value for right "Use". Available are 'grant',
                'deny', 'default' or None
            control (str): value for right "Control". Available are 'grant',
                'deny', 'default' or None
            delete (str): value for right "Delete". Available are 'grant',
                'deny', 'default' or None
            write  (str): value for right "Write". Available are 'grant',
                'deny', 'default' or None
            read (str): value for right "Read". Available are 'grant',
                'deny', 'default' or None
            browse (str): value for right "Browse. Available are 'grant',
                'deny', 'default' or None
        Returns:
            None
        """

        def modify_custom_rights(connection, trustee_id: str, right: Union[Rights, List[Rights]],
                                 to_objects: List[str], object_type: "ObjectTypes", denied: bool,
                                 project: Union[str, "Project"] = None, default: bool = False,
                                 propagate_to_children: Optional[bool] = None,
                                 application: Union[str, "Project"] = None) -> None:
            if application:
                helper.deprecation_warning(
                    '`application`',
                    '`project`',
                    '11.3.4.101',  # NOSONAR
                    False)
                project = project or application
            right_value = _get_custom_right_value(right)
            try:
                _modify_rights(connection=connection, trustee_id=trustee_id, op='REMOVE',
                               rights=right_value, ids=to_objects, object_type=object_type,
                               project=project, denied=(not denied),
                               propagate_to_children=propagate_to_children, verbose=False)
            except HTTPError:
                pass

            op = 'REMOVE' if default else 'ADD'
            verbose = not default
            try:
                _modify_rights(connection=connection, trustee_id=trustee_id, op=op,
                               rights=right_value, ids=to_objects, object_type=object_type,
                               project=project, denied=denied,
                               propagate_to_children=propagate_to_children, verbose=verbose)
            except HTTPError:
                pass

        rights_dict = {
            Rights.EXECUTE: execute,
            Rights.USE: use,
            Rights.CONTROL: control,
            Rights.DELETE: delete,
            Rights.WRITE: write,
            Rights.READ: read,
            Rights.BROWSE: browse
        }
        if not set(rights_dict.values()).issubset({'grant', 'deny', 'default', None}):
            msg = ("Invalid value of the right. Available values are 'grant', 'deny', "
                   "'default' or None.")
            raise ValueError(msg)

        grant_list = [right for right, value in rights_dict.items() if value == 'grant']
        deny_list = [right for right, value in rights_dict.items() if value == 'deny']
        default_list = [right for right, value in rights_dict.items() if value == 'default']

        modify_custom_rights(connection=self.connection, trustee_id=self.id, right=grant_list,
                             to_objects=to_objects, object_type=object_type, denied=False,
                             project=project)
        modify_custom_rights(connection=self.connection, trustee_id=self.id, right=deny_list,
                             to_objects=to_objects, object_type=object_type, denied=True,
                             project=project)
        modify_custom_rights(connection=self.connection, trustee_id=self.id, right=default_list,
                             to_objects=to_objects, object_type=object_type, denied=True,
                             project=project, default=True)


def _modify_rights(connection, trustee_id: str, op: str, rights: int, object_type: "ObjectTypes",
                   ids: List[str], project: Optional[Union[str, "Project"]] = None,
                   propagate_to_children: Optional[bool] = None, denied: Optional[bool] = None,
                   inheritable: Optional[bool] = None, verbose: bool = True) -> None:
    if op not in ["ADD", "REMOVE", "REPLACE"]:
        raise ValueError("Wrong ACL operator passed. Please use ADD, REMOVE or REPLACE")

    # convert Enums
    rights = rights.value if isinstance(rights, Enum) else rights
    if rights not in range(256) and rights not in range(536_870_912, 536_871_168):
        msg = ("Wrong `rights` value, please provide value in range 0-255 or combination of "
               "Rights enums")
        helper.exception_handler(msg)

    if project:
        project = project if isinstance(project, str) else project.id

    if isinstance(ids, str):
        ids = [ids]
    # TODO decide what to do about this code
    for id in ids:
        response = objects.get_object_info(connection=connection, id=id,
                                           object_type=object_type.value, project_id=project)
        if inheritable is None:
            tmp = [
                ace['inheritable']
                for ace in response.json().get('acl', [])
                if ace['trusteeId'] == trustee_id and ace['deny'] == denied
            ]
            inheritable = False if not tmp else tmp[0]

        body = {
            "acl": [{
                "op": op,
                "trustee": trustee_id,
                "rights": rights,
                "type": 1,  # EnumDSSXMLAccessEntryType
                "denied": denied,
                "inheritable": inheritable
            }]
        }

        if isinstance(propagate_to_children, bool):
            body["propagateACLToChildren"] = propagate_to_children

        _ = objects.update_object(connection=connection, id=id, body=body,
                                  object_type=object_type.value, project_id=project,
                                  verbose=verbose)


def _parse_acl_rights_bin_to_dict(rights_bin: int) -> Dict[Rights, bool]:
    # TODO move this to ENUM?
    return {right: rights_bin & right.value != 0 for right in Rights}


def _parse_acl_rights_dict_to_bin(rights_dict: Dict[Rights, bool]) -> int:
    output = 0
    for right, given in rights_dict.items():
        if given:
            output |= right.value
    return output


def _get_custom_right_value(right: Union[Rights, List[Rights]]) -> int:
    # TODO move this to ENUM?
    right_value = 0
    if not isinstance(right, list):
        right = [right]
    for r in right:
        if not isinstance(r, Rights):
            try:
                r = Rights[r.upper()]
            except ValueError:
                msg = (f"Invalid custom `right` value: {r}. Available values are: EXECUTE, USE, "
                       "CONTROL, DELETE, WRITE, READ, BROWSE. See: the Rights enum.")
                raise ValueError(msg)
        right_value |= r.value
    return right_value
